# coding: utf-8
"""
"""
import datetime
from uuid import UUID
import dataclasses
import typing
import re
from flask_babel import _
import flask

from .. import db
from ..models import components, objects, component_authentication
from . import errors
from .utils import cache

# component names are limited to this (semi-arbitrary) length
MAX_COMPONENT_NAME_LENGTH = 100
MAX_COMPONENT_ADDRESS_LENGTH = 100


@dataclasses.dataclass(frozen=True)
class Component:
    """
    This class provides an immutable wrapper around models.federation.FederationComponent.
    """
    id: int
    uuid: str
    name: typing.Optional[str]
    address: typing.Optional[str]
    description: typing.Optional[str]
    last_sync_timestamp: typing.Optional[datetime.datetime]
    import_token_available: bool
    export_token_available: bool
    discoverable: bool
    fed_login_available: bool

    @classmethod
    def from_database(cls, component: components.Component) -> 'Component':
        import_token_available = db.session.query(db.exists().where(component_authentication.OwnComponentAuthentication.component_id == component.id)).scalar()
        export_token_available = db.session.query(db.exists().where(component_authentication.ComponentAuthentication.component_id == component.id)).scalar()
        return Component(
            id=component.id,
            address=component.address,
            uuid=component.uuid,
            name=component.name,
            description=component.description,
            last_sync_timestamp=component.last_sync_timestamp,
            import_token_available=import_token_available,
            export_token_available=export_token_available,
            discoverable=component.discoverable,
            fed_login_available=component.fed_login_available
        )

    def get_name(self) -> str:
        if self.name is None:
            if self.address is not None:
                regex = re.compile(r"https?://(www\.)?")    # should usually be https
                return regex.sub('', self.address).strip().strip('/')
            return _('Database #%(id)s', id=self.id)
        else:
            return self.name

    def update_last_sync_timestamp(
            self,
            last_sync_timestamp: datetime.datetime
    ) -> None:
        component = components.Component.query.filter_by(id=self.id).first()
        if component is None:
            raise errors.ComponentDoesNotExistError()
        component.last_sync_timestamp = last_sync_timestamp
        db.session.add(component)
        db.session.commit()


def validate_address(
        address: str,
        max_length: int = 100,
        allow_http: bool = False
) -> str:
    if not 1 <= len(address) <= max_length:
        raise errors.InvalidComponentAddressError()
    if address[:8] != 'https://' and address[:7] != 'http://':
        address = 'https://' + address

    regex = re.compile(
        r'^(?:http)s?://(?:(?:[A-Z0-9](?:[A-Z0-9-]{0,61}[A-Z0-9])?\.)+(?:[A-Z]{2,6}\.?|[A-Z0-9-]{2,}\.?)|localhost|'
        r'\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})(?::\d+)?(?:/?|[/?]\S+)$', re.IGNORECASE)
    if re.match(regex, address) is None:
        raise errors.InvalidComponentAddressError()

    if not allow_http and address[:7] == 'http://':
        raise errors.InsecureComponentAddressError()

    return address


def get_components() -> typing.List[Component]:
    """
    Returns a list of all existing components.

    :return: the list of components
    """
    return [Component.from_database(component) for component in components.Component.query.order_by(db.asc(components.Component.name)).all()]


def get_federated_login_components() -> typing.List[Component]:
    """
    Returns a list of all components which have the federated login enabled.

    :return: the list of components
    """
    return [Component.from_database(component) for component in components.Component.query.filter_by(fed_login_available=True).order_by(db.asc(components.Component.name)).all()]


def add_component(
        uuid: str,
        name: typing.Optional[str] = None,
        address: typing.Optional[str] = None,
        description: typing.Optional[str] = '',
) -> Component:
    """
    Adds a new component with the given address, name and description.

    :param uuid: the uuid of the component
    :param name: a (possibly empty) name for the component
    :param address: the address of the component
    :param description: a (possibly empty) description for the component
    :return: the newly added component
    :raise errors.InvalidComponentNameError: when the component name is more
        than MAX_COMPONENT_NAME_LENGTH characters long
    :raise errors.InvalidComponentAddressError: when the component address is empty or more
        than MAX_COMPONENT_ADDRESS_LENGTH characters long
    :raise errors.InvalidComponentUUIDError: when the component uuid does not match the UUID format
    :raise errors.ComponentAlreadyExistsError: when another component with the given
        name already exists
    """
    if name is not None and not 1 <= len(name) <= MAX_COMPONENT_NAME_LENGTH:
        raise errors.InvalidComponentNameError()
    if address is not None:
        address = validate_address(address, max_length=MAX_COMPONENT_ADDRESS_LENGTH, allow_http=flask.current_app.config['ALLOW_HTTP'])
    try:
        uuid_obj = UUID(uuid)
        uuid = str(uuid_obj)
    except ValueError:
        raise errors.InvalidComponentUUIDError()
    except TypeError:
        raise errors.InvalidComponentUUIDError()

    if uuid == flask.current_app.config['FEDERATION_UUID'] or name == flask.current_app.config['SERVICE_NAME']:
        raise errors.ComponentAlreadyExistsError(is_current_app=True)

    if name is not None:
        existing_component = components.Component.query.filter_by(name=name).first()
        if existing_component is not None:
            raise errors.ComponentAlreadyExistsError()
    existing_component = components.Component.query.filter_by(uuid=uuid).first()
    if existing_component is not None:
        raise errors.ComponentAlreadyExistsError()

    component = components.Component(uuid=uuid, name=name, description=description, address=address)
    db.session.add(component)
    db.session.commit()

    return Component.from_database(component)


@cache
def check_component_exists(
        component_id: int
) -> None:
    """
    Check whether a component with the given component ID exists.

    :param component_id: the ID of an existing component
    :raise errors.ComponentDoesNotExistError: when no component with the given
        component ID exists
    """
    if not db.session.query(db.exists().where(components.Component.id == component_id)).scalar():
        raise errors.ComponentDoesNotExistError()


def get_component(
        component_id: int
) -> Component:
    """
    Returns the federation component with the given ID.

    :param component_id: the ID of an existing component
    :return: the component
    :raise errors.ComponentDoesNotExistError: when no component with the given
        component ID exists
    """
    component = components.Component.query.filter_by(id=component_id).first()
    if component is None:
        raise errors.ComponentDoesNotExistError()
    return Component.from_database(component)


def get_component_or_none(
        component_id: typing.Optional[int]
) -> typing.Optional[Component]:
    if component_id is None:
        return None
    try:
        return get_component(component_id)
    except errors.ComponentDoesNotExistError:
        return None


def get_component_by_uuid(
        component_uuid: str
) -> Component:
    try:
        uuid_obj = UUID(component_uuid)
        component_uuid = str(uuid_obj)
    except ValueError:
        raise errors.InvalidComponentUUIDError()
    except TypeError:
        raise errors.InvalidComponentUUIDError()
    component = components.Component.query.filter_by(uuid=component_uuid).first()
    if component is None:
        raise errors.ComponentDoesNotExistError
    return Component.from_database(component)


def get_component_id_by_uuid(
        component_uuid: typing.Optional[str]
) -> typing.Optional[int]:
    """
    Get the component ID for a given UUID.

    :param component_uuid: a UUID, or None
    :return: the matching component ID, or None
    """
    if component_uuid is None:
        return None
    # component_uuid is not validated as no exceptions should be raised from
    # this function, so an invalid UUID returns None just as an unknown one.
    component = components.Component.query.filter_by(uuid=component_uuid).first()
    if component is None:
        return None
    return component.id


def update_component(component_id: int, name: typing.Optional[str] = None, address: typing.Optional[str] = None, description: typing.Optional[str] = '') -> None:
    """
    Updates the component's address, name and description.

    :param component_id: the ID of an existing component
    :param name: the new unique component name
    :param address: the address of the component
    :param description: the new component description
    :param name: the new component address
    :raise errors.InvalidComponentNameError: when the component name is more
        than MAX_COMPONENT_NAME_LENGTH characters long
    :raise errors.InvalidComponentAddressError: when the component address is empty or more
        than MAX_COMPONENT_ADDRESS_LENGTH characters long
    :raise errors.ComponentDoesNotExistError: when no component with the given
        component ID exists
    :raise errors.ComponentAlreadyExistsError: when another component with the given
        name or uuid already exists
    """
    if name is not None and not 1 <= len(name) <= MAX_COMPONENT_NAME_LENGTH:
        raise errors.InvalidComponentNameError()
    if address is not None:
        address = validate_address(address, max_length=MAX_COMPONENT_ADDRESS_LENGTH, allow_http=flask.current_app.config['ALLOW_HTTP'])

    component = components.Component.query.filter_by(id=component_id).first()
    if component is None:
        raise errors.ComponentDoesNotExistError()
    if component.name != name and name is not None:
        existing_component = components.Component.query.filter_by(name=name).first()
        if existing_component is not None:
            raise errors.ComponentAlreadyExistsError()

    component.address = address
    component.name = name
    component.description = description or ''
    db.session.add(component)
    db.session.commit()


def get_object_ids_for_component_id(
        component_id: int
) -> typing.Set[int]:
    """
    Get the set of object IDs for all objects imported from a given component.

    :param component_id: the ID of an existing component
    :return: the set of object IDs
    :raise errors.ComponentDoesNotExistError: when no component with the given
        component ID exists
    """
    object_ids = db.session.query(
        objects.Objects._current_table.c.object_id
    ).filter(
        objects.Objects._current_table.c.component_id == component_id
    ).all()
    if not object_ids:
        # ensure component exists
        check_component_exists(component_id)
    return {
        row[0]
        for row in object_ids
    }


def get_object_ids_for_components() -> typing.Set[int]:
    """
    Get the set of object IDs for all objects imported from a component.

    :return: the set of object IDs
    """
    object_ids = db.session.query(
        objects.Objects._current_table.c.object_id
    ).filter(
        db.not_(objects.Objects._current_table.c.component_id.is_(db.null()))
    ).all()
    return {
        row[0]
        for row in object_ids
    }


def get_local_object_ids() -> typing.Set[int]:
    """
    Get the set of object IDs for all objects not imported from any component.

    :return: the set of object IDs
    """
    object_ids = db.session.query(
        objects.Objects._current_table.c.object_id
    ).filter(
        objects.Objects._current_table.c.component_id.is_(db.null())
    ).all()
    return {
        row[0]
        for row in object_ids
    }


@dataclasses.dataclass(frozen=True)
class ComponentInfo:
    """
    This class provides an immutable wrapper around models.components.ComponentInfo.
    """
    id: int
    source_uuid: str
    uuid: str
    name: typing.Optional[str]
    address: typing.Optional[str]
    discoverable: bool
    distance: int

    @classmethod
    def from_database(cls, component_info: components.ComponentInfo) -> 'ComponentInfo':
        return ComponentInfo(
            id=component_info.id,
            address=component_info.address,
            uuid=component_info.uuid,
            name=component_info.name,
            discoverable=component_info.discoverable,
            source_uuid=component_info.source_uuid,
            distance=component_info.distance
        )


def add_or_update_component_info(
        uuid: str,
        source_uuid: str,
        name: typing.Optional[str],
        address: typing.Optional[str],
        discoverable: bool,
        distance: int
) -> ComponentInfo:
    component_info = components.ComponentInfo.query.filter_by(
        uuid=uuid,
        source_uuid=source_uuid
    ).first()
    if component_info is None:
        component_info = components.ComponentInfo(
            uuid=uuid,
            source_uuid=source_uuid,
            name=name,
            address=address,
            discoverable=discoverable,
            distance=distance
        )
    elif distance <= component_info.distance:
        component_info.name = name
        component_info.address = address
        component_info.discoverable = discoverable
        component_info.distance = distance
    db.session.add(component_info)
    db.session.commit()
    return ComponentInfo.from_database(component_info)


def get_component_infos() -> typing.List[ComponentInfo]:
    return [
        ComponentInfo.from_database(component_info)
        for component_info in components.ComponentInfo.query.all()
    ]


def set_component_discoverable(component_id: int, discoverable: bool) -> None:
    """
    Update whether a component should be discoverable.

    :param component_id: the ID of an existing component
    :param discoverable: whether the component should be discoverable
    :raise errors.ComponentDoesNotExistError: when no component with the given
        component ID exists
    """
    component = components.Component.query.filter_by(id=component_id).first()
    if component is None:
        raise errors.ComponentDoesNotExistError()
    component.discoverable = discoverable
    db.session.add(component)
    db.session.commit()


def set_component_fed_login_available(component_id: int, fed_login_available: bool) -> None:
    """
    Update wether the federated login is available for a component.

    :param component_id: the ID of an existing component
    :param fed_login_available: whether the component should be available for the federated login
    :raise errors.ComponentDoesNotExistError: when no component with the given component ID exists
    """
    component = components.Component.query.filter_by(id=component_id).first()
    if component is None:
        raise errors.ComponentDoesNotExistError()
    component.fed_login_available = fed_login_available
    db.session.add(component)
    db.session.commit()



# coding: utf-8
"""

"""
import json
import typing

import flask
import flask_login
from flask_babel import _

from .. import frontend
from ... import logic
from .forms import InviteUserForm, EditGroupForm, LeaveGroupForm, CreateGroupForm, DeleteGroupForm, RemoveGroupMemberForm
from ..users_forms import RevokeInvitationForm
from ...logic.security_tokens import verify_token
from ...logic.languages import get_languages, Language, get_language_by_lang_code, get_language
from ..utils import check_current_user_is_not_readonly
from ...utils import FlaskResponseT
from ...logic.utils import get_translated_text
from ...models import NotificationType


@frontend.route('/groups/', methods=['GET', 'POST'])
@flask_login.login_required
def groups() -> FlaskResponseT:
    user_id = None
    allowed_language_ids = [
        language.id
        for language in get_languages(only_enabled_for_input=True)
    ]
    english = get_language(Language.ENGLISH)
    if 'user_id' in flask.request.args:
        try:
            user_id = int(flask.request.args['user_id'])
        except ValueError:
            pass
    if user_id is not None:
        if user_id != flask_login.current_user.id and not flask_login.current_user.is_admin:
            return flask.abort(403)
        groups = logic.groups.get_user_groups(user_id)
    else:
        groups = logic.groups.get_groups()
    group_membership_by_id = {
        group.id: (flask_login.current_user.id in logic.groups.get_group_member_ids(group.id))
        for group in groups
    }
    create_group_form = CreateGroupForm()
    group_categories = list(logic.group_categories.get_group_categories())
    group_categories.sort(key=lambda category: get_translated_text(category.name).lower())
    create_group_form.categories.choices = [
        (str(category.id), category)
        for category in group_categories
    ]
    show_create_form = False
    if 'create' in flask.request.form:
        check_current_user_is_not_readonly()
        show_create_form = True
        if create_group_form.validate_on_submit():
            if flask_login.current_user.is_admin or not flask.current_app.config['ONLY_ADMINS_CAN_CREATE_GROUPS']:
                try:
                    translations = json.loads(create_group_form.translations.data)
                    if not translations:
                        raise ValueError(_('Please enter at least an english name.'))
                    names = {}
                    descriptions = {}
                    for translation in translations:
                        name = translation['name'].strip()
                        description = translation['description'].strip()
                        language_id = int(translation['language_id'])
                        if language_id not in allowed_language_ids:
                            continue

                        if language_id == Language.ENGLISH:
                            if name == '':
                                raise ValueError(_('Please enter at least an english name.'))

                        lang_code = get_language(language_id).lang_code

                        names[lang_code] = name
                        if description != '':
                            descriptions[lang_code] = description
                        else:
                            descriptions[lang_code] = ''
                    group_id = logic.groups.create_group(names, descriptions, flask_login.current_user.id).id
                    logic.group_categories.set_basic_group_categories(
                        basic_group_id=group_id,
                        category_ids=[
                            int(category_id)
                            for category_id in create_group_form.categories.data
                        ]
                    )
                except ValueError as e:
                    flask.flash(str(e), 'error')
                    create_group_form.translations.errors.append(str(e))
                except logic.errors.GroupAlreadyExistsError:
                    create_group_form.translations.errors.append(_('A basic group with this name already exists.'))
                except logic.errors.InvalidGroupNameError:
                    create_group_form.translations.errors.append(_('This basic group name is invalid.'))
                else:
                    flask.flash(_('The basic group has been created successfully.'), 'success')
                    return flask.redirect(flask.url_for('.group', group_id=group_id))
            else:
                # create_group_form.name.errors.append('Only administrators can create basic groups.')
                create_group_form.translations.errors.append(_('Only administrators can create basic groups.'))
    group_categories_by_id = {
        category.id: category
        for category in group_categories
    }
    group_category_names = logic.group_categories.get_full_group_category_names()
    group_category_tree = logic.group_categories.get_group_category_tree(
        basic_group_ids={group.id for group in groups},
        project_group_ids=set()
    )
    groups.sort(key=lambda group: get_translated_text(group.name).lower())
    return flask.render_template(
        "groups/groups.html",
        groups=groups,
        sorted_by_name=lambda groups: sorted(groups, key=lambda group: get_translated_text(group.name).lower()),
        group_membership_by_id=group_membership_by_id,
        create_group_form=create_group_form,
        show_create_form=show_create_form,
        ENGLISH=english,
        languages=get_languages(only_enabled_for_input=True),
        group_categories=group_categories,
        group_categories_by_id=group_categories_by_id,
        group_category_names=group_category_names,
        group_category_tree=group_category_tree,
    )


@frontend.route('/groups/<int:group_id>', methods=['GET', 'POST'])
@flask_login.login_required
def group(group_id: int) -> FlaskResponseT:
    name_language_ids = []
    description_language_ids = []
    token = flask.request.args.get('token')
    if token:
        expiration_time_limit = flask.current_app.config['INVITATION_TIME_LIMIT']
        token_data = verify_token(token, salt='invite_to_group', secret_key=flask.current_app.config['SECRET_KEY'], expiration=expiration_time_limit)
        if token_data is None:
            flask.flash(_('Invalid basic group invitation token. Please request a new invitation.'), 'error')
            return flask.abort(403)
        if 'invitation_id' in token_data:
            try:
                group_invitation = logic.groups.get_group_invitation(token_data['invitation_id'])
            except logic.errors.GroupInvitationDoesNotExistError:
                flask.flash(_('Unknown basic group invitation. Please request a new invitation.'), 'error')
                return flask.abort(403)
            if group_invitation.accepted:
                flask.flash(_('This invitation token has already been used. Please request a new invitation.'), 'error')
                return flask.abort(403)
            if group_invitation.revoked:
                flask.flash(_('This invitation has been revoked. Please request a new invitation.'), 'error')
                return flask.abort(403)
        if token_data.get('group_id', None) != group_id:
            return flask.abort(403)
        user_id = token_data.get('user_id', None)
        if user_id != flask_login.current_user.id:
            try:
                invited_user = logic.users.get_user(user_id)
                flask.flash(_('Please sign in as user "%(user_name)s" to accept this invitation.', user_name=invited_user.name), 'error')
            except logic.errors.UserDoesNotExistError:
                pass
            return flask.abort(403)

        for notification in logic.notifications.get_notifications(user_id, unread_only=True):
            if notification.type == NotificationType.INVITED_TO_GROUP:
                if notification.data['group_id'] == group_id:
                    logic.notifications.mark_notification_as_read(notification.id)

        try:
            logic.groups.add_user_to_group(group_id, user_id)
        except logic.errors.UserAlreadyMemberOfGroupError:
            flask.flash(_('You are already a member of this basic group.'), 'error')
        except logic.errors.GroupDoesNotExistError:
            pass
    try:
        group_member_ids = logic.groups.get_group_member_ids(group_id)
    except logic.errors.GroupDoesNotExistError:
        flask.flash(_('This basic group does not exist.'), 'error')
        return flask.abort(404)
    group_member_ids.sort(key=lambda user_id: (logic.users.get_user(user_id).name or '').lower())
    user_is_member = flask_login.current_user.id in group_member_ids
    user_can_leave = user_is_member
    # admins are treated as members for the sake of permissions
    user_is_member = user_is_member or flask_login.current_user.has_admin_permissions
    group = logic.groups.get_group(group_id)
    show_edit_form = False
    revoke_invitation_form = None

    english = get_language(Language.ENGLISH)
    allowed_language_ids = [
        language.id
        for language in get_languages(only_enabled_for_input=True)
    ]

    class GroupTranslation(typing.TypedDict):
        language_id: int
        lang_name: str
        name: str
        description: str

    translations: typing.List[GroupTranslation] = []
    if user_is_member or flask_login.current_user.has_admin_permissions:
        show_objects_link = True
        if user_can_leave:
            leave_group_form = LeaveGroupForm()
        else:
            leave_group_form = None
        invite_user_form = InviteUserForm()
        revoke_invitation_form = RevokeInvitationForm()
        if 'revoke_invitation' in flask.request.form and revoke_invitation_form.validate_on_submit():
            invitation_id = revoke_invitation_form.invitation_id.data
            try:
                invitation = logic.groups.get_group_invitation(invitation_id)
            except logic.errors.GroupInvitationDoesNotExistError:
                flask.flash(_('Unknown basic group invitation.'), 'error')
                return flask.redirect(flask.url_for('.group', group_id=group_id))
            if invitation.accepted:
                flask.flash(_('This basic group invitation has already been accepted.'), 'error')
            else:
                logic.groups.revoke_group_invitation(invitation_id)
                flask.flash(_('The invitation has been revoked.'), 'success')
            return flask.redirect(flask.url_for('.group', group_id=group_id))
        edit_group_form = EditGroupForm()
        group_categories = list(logic.group_categories.get_group_categories())
        group_categories.sort(key=lambda category: get_translated_text(category.name).lower())
        edit_group_form.categories.choices = [
            (str(category.id), category)
            for category in group_categories
        ]
        if flask_login.current_user.is_admin or not flask.current_app.config['ONLY_ADMINS_CAN_DELETE_GROUPS'] or set(group_member_ids) == {flask_login.current_user.id}:
            delete_group_form = DeleteGroupForm()
        else:
            delete_group_form = None
        remove_group_member_form = RemoveGroupMemberForm()

        for lang_code, name in group.name.items():
            lang_id = get_language_by_lang_code(lang_code).id
            name_language_ids.append(lang_id)
            translation = GroupTranslation(
                language_id=lang_id,
                lang_name=get_translated_text(get_language(lang_id).names),
                name=name,
                description=''
            )
            translations.append(translation)

        for lang_code, description in group.description.items():
            if lang_code == '':
                continue
            found = False
            lang_id = get_language_by_lang_code(lang_code).id
            description_language_ids.append(lang_id)
            for translation in translations:
                if lang_id == translation['language_id']:
                    translation['description'] = description
                    found = True
                    break
            if not found:
                translation = GroupTranslation(
                    language_id=lang_id,
                    lang_name=get_translated_text(get_language(lang_id).names),
                    name='',
                    description=description[1]
                )
                translations.append(translation)

        if 'edit' in flask.request.form:
            check_current_user_is_not_readonly()
            show_edit_form = True
            if edit_group_form.validate_on_submit():
                try:
                    translations = json.loads(edit_group_form.translations.data)
                    if not translations:
                        raise ValueError(_('Please enter at least an english name.'))
                    names = {}
                    descriptions = {}
                    for translation in translations:
                        name = translation['name'].strip()
                        description = translation['description'].strip()
                        language_id = int(translation['language_id'])

                        if language_id not in allowed_language_ids:
                            continue

                        if language_id == Language.ENGLISH:
                            if name == '':
                                raise ValueError(_('Please enter at least an english name.'))
                        elif name == '' and description == '':
                            continue

                        lang_code = get_language(language_id).lang_code

                        names[lang_code] = name
                        if description != '':
                            descriptions[lang_code] = description
                        else:
                            descriptions[lang_code] = ''
                    logic.groups.update_group(group_id, names, descriptions)
                    logic.group_categories.set_basic_group_categories(
                        basic_group_id=group_id,
                        category_ids=[
                            int(category_id)
                            for category_id in edit_group_form.categories.data
                        ]
                    )
                except ValueError as e:
                    flask.flash(str(e), 'error')
                    edit_group_form.translations.errors.append(str(e))
                except logic.errors.GroupDoesNotExistError:
                    flask.flash(_('This basic group does not exist.'), 'error')
                    return flask.redirect(flask.url_for('.groups'))
                except logic.errors.GroupAlreadyExistsError:
                    edit_group_form.translations.errors.append(_('A basic group with this name already exists.'))
                except logic.errors.InvalidGroupNameError:
                    edit_group_form.translations.errors.append(_('This basic group name is invalid.'))
                else:
                    flask.flash(_('Basic group information updated successfully.'), 'success')
                    return flask.redirect(flask.url_for('.group', group_id=group_id))
        elif 'add_user' in flask.request.form:
            check_current_user_is_not_readonly()
            if invite_user_form.validate_on_submit():
                try:
                    if invite_user_form.add_directly.data and flask_login.current_user.is_admin:
                        logic.groups.add_user_to_group(
                            group_id=group_id,
                            user_id=invite_user_form.user_id.data
                        )
                        flask.flash(_('The user was successfully added to the basic group.'), 'success')
                    else:
                        logic.groups.invite_user_to_group(
                            group_id=group_id,
                            user_id=invite_user_form.user_id.data,
                            inviter_id=flask_login.current_user.id
                        )
                        flask.flash(_('The user was successfully invited to the basic group.'), 'success')
                except logic.errors.GroupDoesNotExistError:
                    flask.flash(_('This basic group does not exist.'), 'error')
                    return flask.redirect(flask.url_for('.groups'))
                except logic.errors.UserDoesNotExistError:
                    flask.flash(_('This user does not exist.'), 'error')
                except logic.errors.UserAlreadyMemberOfGroupError:
                    flask.flash(_('This user is already a member of this basic group.'), 'error')
                else:
                    return flask.redirect(flask.url_for('.group', group_id=group_id))
        elif 'leave' in flask.request.form and leave_group_form is not None:
            if user_can_leave and leave_group_form.validate_on_submit():
                try:
                    logic.groups.remove_user_from_group(group_id, flask_login.current_user.id)
                except logic.errors.GroupDoesNotExistError:
                    flask.flash(_('This basic group does not exist.'), 'error')
                    return flask.redirect(flask.url_for('.groups'))
                except logic.errors.UserDoesNotExistError:
                    return flask.abort(500)
                except logic.errors.UserNotMemberOfGroupError:
                    flask.flash(_('You have already left the basic group.'), 'error')
                    return flask.redirect(flask.url_for('.groups'))
                else:
                    flask.flash(_('You have successfully left the basic group.'), 'success')
                    return flask.redirect(flask.url_for('.groups'))
        elif 'delete' in flask.request.form and delete_group_form:
            check_current_user_is_not_readonly()
            if delete_group_form.validate_on_submit():
                try:
                    logic.groups.delete_group(group_id)
                except logic.errors.GroupDoesNotExistError:
                    flask.flash(_('This basic group has already been deleted.'), 'success')
                    return flask.redirect(flask.url_for('.groups'))
                else:
                    flask.flash(_('You have successfully deleted the basic group.'), 'success')
                    return flask.redirect(flask.url_for('.groups'))
        elif 'remove_member' in flask.request.form:
            check_current_user_is_not_readonly()
            if remove_group_member_form.validate_on_submit():
                member_id_str = flask.request.form['remove_member']
                try:
                    member_id = int(member_id_str)
                except ValueError:
                    flask.flash(_('The member ID was invalid. Please contact an administrator.'), 'error')
                    return flask.redirect(flask.url_for('.group', group_id=group_id))
                try:
                    logic.groups.remove_user_from_group(group_id, member_id)
                except logic.errors.GroupDoesNotExistError:
                    flask.flash(_('This basic group does not exist.'), 'error')
                    return flask.redirect(flask.url_for('.groups'))
                except logic.errors.UserDoesNotExistError:
                    flask.flash(_('This user does not exist.'), 'error')
                    return flask.redirect(flask.url_for('.group', group_id=group_id))
                except logic.errors.UserNotMemberOfGroupError:
                    flask.flash(_('This user is not a member of this basic group.'), 'success')
                    return flask.redirect(flask.url_for('.group', group_id=group_id))
                else:
                    flask.flash(_('This user has been removed from this basic group.'), 'success')
                    return flask.redirect(flask.url_for('.group', group_id=group_id))
    else:
        if flask.request.method.lower() == 'post':
            return flask.abort(403)
        leave_group_form = None
        edit_group_form = None
        invite_user_form = None
        delete_group_form = None
        remove_group_member_form = None
        show_objects_link = False
        group_categories = None

    group_invitations = None
    show_invitation_log = flask_login.current_user.is_admin and logic.settings.get_user_setting(flask_login.current_user.id, 'SHOW_INVITATION_LOG')
    if user_is_member or flask_login.current_user.is_admin:
        group_invitations = logic.groups.get_group_invitations(
            group_id=group_id,
            include_accepted_invitations=show_invitation_log,
            include_expired_invitations=show_invitation_log,
            include_revoked_invitations=show_invitation_log
        )

    if english.id not in description_language_ids:
        description_language_ids.append(english.id)
    basic_group_categories = logic.group_categories.get_basic_group_categories(group.id)
    category_names = logic.group_categories.get_full_group_category_names()
    return flask.render_template(
        'groups/group.html',
        ENGLISH=english,
        translations=translations,
        languages=get_languages(only_enabled_for_input=True),
        name_language_ids=name_language_ids,
        description_language_ids=description_language_ids,
        group=group,
        group_member_ids=group_member_ids,
        group_categories=group_categories,
        group_category_names=logic.group_categories.get_full_group_category_names(),
        basic_group_categories=basic_group_categories,
        category_names=category_names,
        get_users=logic.users.get_users,
        get_user=logic.users.get_user,
        group_invitations=group_invitations,
        show_objects_link=show_objects_link,
        show_invitation_log=show_invitation_log,
        leave_group_form=leave_group_form,
        delete_group_form=delete_group_form,
        remove_group_member_form=remove_group_member_form,
        edit_group_form=edit_group_form,
        invite_user_form=invite_user_form,
        show_edit_form=show_edit_form,
        revoke_invitation_form=revoke_invitation_form
    )



"""
This module offers functions for exporting SampleDB information to a Dataverse
instance with support for the Processing Metadata from the EngMeta schemas
(see: https://www.izus.uni-stuttgart.de/fokus/engmeta).

Some basic information is exported to the default Citation Metadata block and
the process-specific metadata is exported to the Method Parameters fields of
the Processing Metadata block.
"""

import datetime
import json
import typing
from urllib.parse import urlencode

import flask
import requests

from .components import get_component_by_uuid
from .. import db
from .units import prettify_units
from .utils import get_translated_text
from . import actions, datatypes, object_log, users, objects, errors, object_permissions, files, settings
from .schemas.utils import data_iter
from ..models import DataverseExport, DataverseExportStatus, Permissions, ObjectLogEntryType


DATAVERSE_TIMEOUT = 30


def flatten_metadata(
        metadata: typing.Union[typing.Dict[str, typing.Any], typing.List[typing.Any]]
) -> typing.Generator[typing.Tuple[typing.Dict[str, typing.Any], typing.List[typing.Union[str, int]], typing.List[typing.Union[str, int]]], None, None]:
    """
    Convert nested object data to a generator yielding the property data and path.

    :param metadata: the nested object data
    :return: the generator for property data, whitelist path and title path
    """
    for property_path, property_data in data_iter(metadata):
        if not isinstance(property_data, dict) or '_type' not in property_data:
            continue
        if property_data['_type'] == 'timeseries':
            for row in property_data['data']:
                utc_datetime_string, magnitude = row[:2]
                row_metadata = {
                    '_type': 'quantity',
                    'units': property_data['units'],
                    'dimensionality': property_data['dimensionality'],
                    'magnitude': magnitude
                }
                yield row_metadata, list(property_path), list(property_path) + [utc_datetime_string]
        else:
            yield property_data, list(property_path), list(property_path)


def _translations_to_str(
        content: typing.Optional[typing.Union[str, typing.Dict[str, str]]]
) -> str:
    return get_translated_text(content, language_code='en')


def get_title_for_property(
        path: typing.List[typing.Union[str, int]],
        schema: typing.Dict[str, typing.Any]
) -> str:
    """
    Return the title path for a given property.

    The title path consists of the individual titles or indices, joined by the
    rightwards arrow symbol.

    :param path: the path to a property
    :param schema: the object schema
    :return: the title path
    """
    try:
        subschema = schema
        title_path = []
        for key in path:
            if isinstance(key, str) and 'properties' in subschema:
                subschema = subschema['properties'][key]
                if isinstance(subschema['title'], dict):
                    title_path.append(subschema['title'].get('en', subschema['title']))
                else:
                    title_path.append(subschema['title'])
            elif isinstance(key, int) and 'items' in subschema:
                subschema = subschema['items']
                title_path.append(key)
            else:
                # fallback to just appending the keys to the title path
                subschema = {}
                title_path.append(key)
    except Exception:
        title_path = path
    return ' → '.join(map(str, title_path))


def get_property_export_default(
        path: typing.List[str],
        schema: typing.Dict[str, typing.Any]
) -> bool:
    """
    Return whether dataverse_export is set to True in the schema of a property.

    :param path: the path to a property
    :param schema: the object schema
    :return: the value of dataverse_export, or False if it is not set
    """
    if path == ['name']:
        return True
    try:
        subschema = schema
        for key in path:
            if isinstance(key, str):
                subschema = subschema['properties'][key]
            else:
                subschema = subschema['items']
        return bool(subschema.get('dataverse_export', False))
    except Exception:
        return False


def _convert_metadata_to_process(
        metadata: typing.Dict[str, typing.Any],
        schema: typing.Dict[str, typing.Any],
        user_id: int,
        property_whitelist: typing.Iterable[typing.Sequence[typing.Union[str, int]]],
        object_id: int
) -> typing.List[typing.Dict[str, typing.Any]]:
    fields = []
    property_whitelist = {
        tuple(whitelist_path)
        for whitelist_path in property_whitelist
    }
    for value, whitelist_path, title_path in flatten_metadata(metadata):
        if tuple(whitelist_path) not in property_whitelist:
            continue
        title = get_title_for_property(title_path, schema)
        units = ''
        magnitude = ''
        text_value = ''
        symbol = ''
        if value['_type'] == 'quantity':
            units = prettify_units(value['units'])
            if units == '1':
                # dimensionless quantities with a factor of 1 do not need units
                units = ''
            magnitude = f"{float(datatypes.Quantity.from_json(value).magnitude):g}"
        elif value['_type'] == 'text':
            text_value = _translations_to_str(datatypes.Text.from_json(value).text) or ''
        elif value['_type'] == 'bool':
            text_value = str(datatypes.Boolean.from_json(value).value)
        elif value['_type'] == 'datetime':
            text_value = datatypes.DateTime.from_json(value).utc_datetime.isoformat(timespec='microseconds')
        elif value['_type'] == 'hazards':
            hazard_names = {
                1: 'Explosive',
                2: 'Flammable',
                3: 'Oxidizing',
                4: 'Compressed Gas',
                5: 'Corrosive',
                6: 'Toxic',
                7: 'Harmful',
                8: 'Health Hazard',
                9: 'Environmental Hazard'
            }
            text_value = ', '.join(hazard_names[hazard_id] for hazard_id in sorted(value['hazards']))
        elif value['_type'] == 'tags':
            # tags are handled separately via the keywords in Citation Metadata
            continue
        elif value['_type'] == 'file':
            try:
                file = files.get_file(value['file_id'], object_id)
                if file.is_hidden:
                    text_value = f'Hidden (#{value["file_id"]})'
                else:
                    text_value = file.original_file_name
            except errors.FileDoesNotExistError:
                text_value = f'Unknown (#{value["file_id"]})'
            except errors.InvalidFileStorageError:
                text_value = f'Unnamed (#{value["file_id"]})'
        elif value['_type'] == 'user':
            if 'component_uuid' in value:
                try:
                    component = get_component_by_uuid(value['component_uuid'])
                    try:
                        name = users.get_user(value['user_id'], component.id).name
                        if name is None:
                            text_value = f'Unknown (#{value["user_id"]} @ {component.get_name()})'
                        else:
                            text_value = f'{name} (#{value["user_id"]} @ {component.get_name()})'
                    except errors.UserDoesNotExistError:
                        text_value = f'Unknown (#{value["user_id"]} @ {component.get_name()})'
                except errors.ComponentDoesNotExistError:
                    text_value = f'Unknown (#{value["user_id"]} @ {value["component_uuid"]})'
            else:
                try:
                    user = users.get_user(value["user_id"])
                    if user.name is None:
                        name = 'Unknown'
                    else:
                        name = user.name
                    if user.component is not None:
                        text_value = f'{name} (#{value["user_id"]}, #{user.fed_id} @ {user.component.get_name()}))'
                    else:
                        text_value = f'{name} (#{value["user_id"]})'
                except errors.UserDoesNotExistError:
                    text_value = f'Unknown (#{value["user_id"]})'
        elif value['_type'] in ('sample', 'measurement', 'object_reference'):
            object_id = value['object_id']
            if 'component_uuid' in value:
                try:
                    component = get_component_by_uuid(value['component_uuid'])
                    component_id = component.id
                    component_name = component.get_name()
                except errors.ComponentDoesNotExistError:
                    component_id = None
                    component_name = value['component_uuid']
            else:
                component_name = None
                component_id = None
            try:
                if component_id:
                    obj = objects.get_fed_object(object_id, component_id)
                else:
                    obj = objects.get_object(object_id)
                if Permissions.READ in object_permissions.get_user_object_permissions(object_id, user_id) and obj.data:
                    object_name = _translations_to_str(obj.data.get('name', {}).get('text'))
                else:
                    object_name = None
                if obj.component:
                    component_name = obj.component.get_name()
                    fed_id = obj.fed_object_id
                else:
                    component_name = None
                    fed_id = obj.fed_object_id
            except errors.ObjectDoesNotExistError:
                fed_id = value['object_id']
                object_name = None
            if object_name:
                if component_name:
                    text_value = f"{object_name} (#{object_id}, #{fed_id} @ {component_name})"
                else:
                    text_value = f"{object_name} (#{object_id})"
            else:
                if component_name:
                    text_value = f"#{object_id} (#{fed_id} @ {component_name})"
                else:
                    text_value = f"#{object_id}"
        else:
            continue
        fields.append({
            'processMethodsParName': {
                'typeName': 'processMethodsParName',
                'multiple': False,
                'typeClass': 'primitive',
                'value': title
            },
            'processMethodsParUnit': {
                'typeName': 'processMethodsParUnit',
                'multiple': False,
                'typeClass': 'primitive',
                'value': units
            },
            'processMethodsParValue': {
                'typeName': 'processMethodsParValue',
                'multiple': False,
                'typeClass': 'primitive',
                'value': magnitude
            },
            'processMethodsParTextValue': {
                'typeName': 'processMethodsParTextValue',
                'multiple': False,
                'typeClass': 'primitive',
                'value': text_value
            },
            'processMethodsParSymbol': {
                'typeName': 'processMethodsParSymbol',
                'multiple': False,
                'typeClass': 'primitive',
                'value': symbol
            }
        })
    return fields


def upload_object(
        object_id: int,
        user_id: int,
        server_url: str,
        api_token: str,
        dataverse: str,
        property_whitelist: typing.Optional[typing.Sequence[typing.Sequence[typing.Union[str, int]]]] = None,
        file_id_whitelist: typing.Sequence[int] = (),
        tag_whitelist: typing.Sequence[str] = ()
) -> typing.Tuple[bool, str]:
    """
    Uploads object information and files to a given Dataverse.

    Only properties and non-hidden files included in the provided whitelists
    will be exported, aside from the name and auxiliary information such as
    the deposition date and the list of authors.

    :param object_id: the ID of an existing object to upload
    :param user_id: the ID of the uploading user
    :param server_url: the base URL of a Dataverse server
    :param api_token: an API token for the Dataverse instance
    :param dataverse: the Dataverse to upload to
    :param property_whitelist: a list of property paths to be uploaded
    :param file_id_whitelist: a list of file IDs to be uploaded
    :param tag_whitelist: a list of tags to be set as keywords
    :return: whether the upload was successful and the Dataverse url
        or a dictionary containing additional information
    :raise DataverseNotReachableError: when an HTTP request to the Dataverse
        instance fails
    """
    if property_whitelist is None:
        property_whitelist = [['name']]

    object = objects.get_object(object_id)

    depositor = users.get_user(user_id)
    date_of_deposit = datetime.date.today()

    sampledb_metadata = object.data
    schema = object.schema

    instrument_name = None
    if object.action_id and not flask.current_app.config['DISABLE_INSTRUMENTS']:
        action = actions.get_action(object.action_id)
        if action.instrument:
            instrument_name = action.instrument.name.get('en', 'Unnamed Instrument')

    author_ids = set()
    for entry in object_log.get_object_log_entries(object.id, user_id):
        if entry.type in {
            ObjectLogEntryType.CREATE_BATCH,
            ObjectLogEntryType.CREATE_OBJECT,
            ObjectLogEntryType.EDIT_OBJECT,
            ObjectLogEntryType.RESTORE_OBJECT_VERSION,
        }:
            author_ids.add(entry.user_id)

    authors = [
        users.get_user(author_id)
        for author_id in author_ids
    ]
    authors.sort(key=lambda author: author.name)  # type: ignore

    tags = set()
    if object.data:
        for property in object.data.values():
            if '_type' in property and property['_type'] == 'tags':
                for tag in property['tags']:
                    if tag in tag_whitelist:
                        tags.add(tag)
    sorted_tags = list(tags)
    sorted_tags.sort()

    author_metadata = []

    for author in authors:
        author_metadata.append({
            'authorName': {
                'typeName': 'authorName',
                'multiple': False,
                'typeClass': 'primitive',
                'value': author.name
            }
        })
        if author.orcid:
            author_metadata[-1]["authorIdentifierScheme"] = {
                "typeName": "authorIdentifierScheme",
                "multiple": False,
                "typeClass": "controlledVocabulary",
                "value": "ORCID"
            }
            author_metadata[-1]["authorIdentifier"] = {
                "typeName": "authorIdentifier",
                "multiple": False,
                "typeClass": "primitive",
                "value": author.orcid
            }
        if author.affiliation:
            author_metadata[-1]["authorAffiliation"] = {
                "typeName": "authorAffiliation",
                "multiple": False,
                "typeClass": "primitive",
                "value": author.affiliation
            }

    if sampledb_metadata is not None and schema is not None:
        method_parameters = _convert_metadata_to_process(sampledb_metadata, schema, user_id, property_whitelist, object_id)
    else:
        method_parameters = []

    if object.component is None or object.component.uuid == flask.current_app.config['FEDERATION_UUID']:
        description = f'Dataset exported from {flask.current_app.config["SERVICE_NAME"]}.'
    else:
        object_component = object.component
        description = f'Dataset exported as object #{object.object_id} from {flask.current_app.config["SERVICE_NAME"]} and created as object #{object.fed_object_id} at {object_component.get_name()}.'

    citation_metadata = {
        'displayName': 'Citation Metadata',
        'fields': [
            {
                'typeName': 'title',
                'multiple': False,
                'typeClass': 'primitive',
                'value': (sampledb_metadata['name']['text'].get('en', json.dumps(sampledb_metadata['name']['text'])) if isinstance(sampledb_metadata['name']['text'], dict) else json.dumps(sampledb_metadata['name']['text'])) if sampledb_metadata is not None else ''
            },
            {
                'typeName': 'alternativeURL',
                'multiple': False,
                'typeClass': 'primitive',
                'value': flask.url_for('frontend.object', object_id=object_id, _external=True)
            },
            {
                'typeName': 'otherId',
                'multiple': True,
                'typeClass': 'compound',
                'value': [
                    {
                        'otherIdAgency': {
                            'typeName': 'otherIdAgency',
                            'multiple': False,
                            'typeClass': 'primitive',
                            'value': flask.current_app.config['SERVICE_NAME']
                        },
                        'otherIdValue': {
                            'typeName': 'otherIdValue',
                            'multiple': False,
                            'typeClass': 'primitive',
                            'value': str(object_id)
                        }
                    }
                ]
            },
            {
                'typeName': 'author',
                'multiple': True,
                'typeClass': 'compound',
                'value': author_metadata
            },
            {
                'typeName': 'datasetContact',
                'multiple': True,
                'typeClass': 'compound',
                'value': [
                    {
                        'datasetContactName': {
                            'typeName': 'datasetContactName',
                            'multiple': False,
                            'typeClass': 'primitive',
                            'value': depositor.name
                        },
                        'datasetContactEmail': {
                            'typeName': 'datasetContactEmail',
                            'multiple': False,
                            'typeClass': 'primitive',
                            'value': depositor.email
                        }
                    }
                ]
            },
            {
                'typeName': 'dsDescription',
                'multiple': True,
                'typeClass': 'compound',
                'value': [
                    {
                        'dsDescriptionValue': {
                            'typeName': 'dsDescriptionValue',
                            'multiple': False,
                            'typeClass': 'primitive',
                            'value': description
                        }
                    }
                ]
            },
            {
                'typeName': 'subject',
                'multiple': True,
                'typeClass': 'controlledVocabulary',
                'value': ['Other']
            },
            {
                'typeName': 'depositor',
                'multiple': False,
                'typeClass': 'primitive',
                'value': depositor.name
            },
            {
                'typeName': 'dateOfDeposit',
                'multiple': False,
                'typeClass': 'primitive',
                'value': date_of_deposit.strftime('%Y-%m-%d')
            }
        ]
    }

    if object.component is not None:
        citation_metadata['fields'].append({  # type: ignore
            'typeName': 'otherId',
            'multiple': True,
            'typeClass': 'compound',
            'value': [
                {
                    'otherIdAgency': {
                        'typeName': 'otherIdAgency',
                        'multiple': False,
                        'typeClass': 'primitive',
                        'value': object.component.get_name()
                    },
                    'otherIdValue': {
                        'typeName': 'otherIdValue',
                        'multiple': False,
                        'typeClass': 'primitive',
                        'value': str(object.fed_object_id)
                    }
                }
            ]
        })

    if sorted_tags:
        citation_metadata['fields'].append({  # type: ignore
            "typeName": "keyword",
            "multiple": True,
            "typeClass": "compound",
            "value": [
                {
                    "keywordValue": {
                        "typeName": "keywordValue",
                        "multiple": False,
                        "typeClass": "primitive",
                        "value": tag
                    }
                } for tag in sorted_tags
            ]
        })

    process_metadata = {
        'displayName': 'Process Metadata',
        'fields': []
    }

    if method_parameters:
        process_metadata['fields'].append({  # type: ignore
            'typeName': 'processMethodsPar',
            'multiple': True,
            'typeClass': 'compound',
            'value': method_parameters
        })

    if instrument_name:
        process_metadata['fields'].append({  # type: ignore
            "typeName": "processInstru",
            "multiple": True,
            "typeClass": "compound",
            "value": [
                {
                    "processInstruName": {
                        "typeName": "processInstruName",
                        "multiple": False,
                        "typeClass": "primitive",
                        "value": instrument_name
                    }
                }
            ]
        })

    metadata_blocks = {
        'citation': citation_metadata,
        'process': process_metadata
    }

    dataverse_metadata = {
        'datasetVersion': {
            'metadataBlocks': metadata_blocks
        }
    }

    try:
        r = requests.post(
            f'{server_url}/api/v1/dataverses/{dataverse}/datasets/',
            headers={
                'X-Dataverse-key': api_token
            },
            allow_redirects=False,
            json=dataverse_metadata,
            timeout=DATAVERSE_TIMEOUT
        )
    except requests.exceptions.RequestException:
        raise errors.DataverseNotReachableError()
    try:
        result = r.json()
        if r.status_code == 201 and result['status'] == 'OK':
            persistent_id = result['data']['persistentId']
            dataverse_url = f'{server_url}/dataset.xhtml?{urlencode({"persistentId": persistent_id})}'
            dataverse_export = DataverseExport.query.filter_by(object_id=object_id).first()
            if dataverse_export and dataverse_export.status == DataverseExportStatus.TASK_CREATED:
                DataverseExport.query.filter_by(object_id=object_id).update({
                    'status': DataverseExportStatus.EXPORT_FINISHED,
                    'dataverse_url': dataverse_url
                })
            else:
                dataverse_export = DataverseExport(object_id, dataverse_url, user_id, DataverseExportStatus.EXPORT_FINISHED)
                db.session.add(dataverse_export)
            db.session.commit()
            object_log.export_to_dataverse(user_id, object_id, dataverse_url)
            _upload_files_to_dataset(server_url, api_token, persistent_id, object_id, file_id_whitelist)
            return True, dataverse_url
        return False, r.json()['message']
    except Exception:
        return False, ""


def _upload_files_to_dataset(
    server_url: str,
    api_token: str,
    persistent_id: str,
    object_id: int,
    file_id_whitelist: typing.Sequence[int]
) -> None:
    """
    Upload files for an object to the given Dataverse dataset.

    Only non-hidden files with IDs from the file_id_whitelist and in local
    storage will be uploaded. Failed file uploads will be ignored

    :param server_url: the base URL of a Dataverse server
    :param api_token: an API token for the Dataverse instance
    :param persistent_id: the Dataverse ID of the dataset
    :param object_id: the ID of the corresponding object
    :param file_id_whitelist: a list of file IDs to be uploaded
    :raise errors.DataverseNotReachableError: if there was an error during
        communication with the Dataverse API
    """
    for file in files.get_files_for_object(object_id):
        if file.id not in file_id_whitelist:
            continue
        if file.is_hidden:
            continue
        if file.storage == 'database':
            file_name = file.original_file_name
            file_content = file.open(read_only=True).read()
            if file.title and file.description:
                file_description = file.title + '\n\n' + file.description
            elif file.title:
                file_description = file.title
            elif file.description:
                file_description = file.description
            else:
                file_description = ''
            try:
                requests.post(
                    f'{server_url}/api/datasets/:persistentId/add',
                    data={
                        'jsonData': json.dumps({
                            'description': file_description,
                            'categories': []
                        })
                    },
                    files={
                        'file': (file_name, file_content),
                    },
                    params={
                        'persistentId': persistent_id
                    },
                    headers={
                        'X-Dataverse-key': api_token
                    },
                    timeout=DATAVERSE_TIMEOUT
                )
            except requests.exceptions.RequestException:
                # ignore failed file uploads
                pass


def list_dataverses(
        server_url: str,
        api_token: str,
        dataverse: str
) -> typing.List[typing.Tuple[int, typing.Dict[str, typing.Any]]]:
    """
    Return information on a dataverse and its children.

    For each dataverse, the list returned by this function contains the degree
    by which it is descended from the given dataverse, and a dictionary
    containing their ID and title.

    :param server_url: the base URL of a Dataverse server
    :param api_token: an API token for the Dataverse instance
    :param dataverse: the Dataverse to get information for
    :return: a list with information on dataverses
    :raise errors.DataverseNotReachableError: if there was an error during
        communication with the Dataverse API
    """
    dataverses = []

    dataverses_with_children = [
        (1, dataverse)
        for dataverse in _list_dataverses_nested(server_url, api_token, dataverse)
    ]

    while dataverses_with_children:
        level, dataverse_with_children = dataverses_with_children.pop(0)
        dataverses.append((level, {
            'id': dataverse_with_children.get('id'),
            'title': dataverse_with_children.get('title')
        }))
        for child_dataverse in reversed(dataverse_with_children['child_dataverses']):
            dataverses_with_children.insert(0, (level + 1, child_dataverse))

    return dataverses


def get_dataverse_info(
        server_url: str,
        api_token: str,
        dataverse: str
) -> typing.Optional[typing.Dict[str, typing.Any]]:
    """
    Return information on a given Dataverse.

    :param server_url: the base URL of a Dataverse server
    :param api_token: an API token for the Dataverse instance
    :param dataverse: the Dataverse to get information for
    :return: a dictionary containing the Dataverse ID and title, or None
    :raise errors.DataverseNotReachableError: if there was an error during
        communication with the Dataverse API
    """
    try:
        r = requests.get(
            f'{server_url}/api/v1/dataverses/{dataverse}',
            headers={
                'X-Dataverse-key': api_token
            },
            allow_redirects=False,
            timeout=DATAVERSE_TIMEOUT
        )
    except requests.exceptions.RequestException:
        raise errors.DataverseNotReachableError()
    if r.status_code != 200:
        return None
    try:
        result = r.json()
        if result.get('status') != 'OK':
            return None
        return {
            'id': result['data']['id'],
            'title': result['data']['name']
        }
    except Exception:
        return None


def _list_dataverses_nested(
        server_url: str,
        api_token: str,
        dataverse: str
) -> typing.List[typing.Dict[str, typing.Any]]:
    """
    Return information on Dataverses within a given Dataverse.

    :param server_url: the base URL of a Dataverse server
    :param api_token: an API token for the Dataverse instance
    :param dataverse: the root Dataverse to search in
    :return: a list of dictionaries containing the Dataverse information
    :raise errors.DataverseNotReachableError: if there was an error during
        communication with the Dataverse API
    """
    if dataverse is None:
        return []
    try:
        r = requests.get(
            f'{server_url}/api/v1/dataverses/{dataverse}/contents',
            headers={
                'X-Dataverse-key': api_token
            },
            allow_redirects=False,
            timeout=DATAVERSE_TIMEOUT
        )
    except requests.exceptions.RequestException:
        raise errors.DataverseNotReachableError()
    if r.status_code != 200:
        return []
    try:
        result = r.json()
    except Exception:
        return []
    if result.get('status') != 'OK':
        return []
    child_dataverses = []
    for entry in result.get('data', ()):
        if entry.get('type') == "dataverse":
            if 'id' not in entry or 'title' not in entry:
                continue
            child_dataverses.append({
                'id': entry.get('id'),
                'title': entry.get('title'),
                'child_dataverses': _list_dataverses_nested(server_url, api_token, entry.get('id'))
            })
    return child_dataverses


def get_dataverse_export_state(object_id: int) -> typing.Optional[DataverseExportStatus]:
    """
    Returns the current status of a dataverse export.

    :param object_id: the ID of an existing object
    :return: the current state or None
    """
    dataverse_export: typing.Optional[DataverseExport] = DataverseExport.query.filter_by(object_id=object_id).first()
    if dataverse_export is None:
        return None
    return dataverse_export.status


def get_dataverse_url(object_id: int) -> typing.Optional[str]:
    """
    Return the URL for an object in a Dataverse instance.

    :param object_id: the ID of an existing object
    :return: the URL or None
    """
    dataverse_export: typing.Optional[DataverseExport] = DataverseExport.query.filter_by(object_id=object_id).first()
    if dataverse_export is None:
        return None
    return dataverse_export.dataverse_url


def get_user_valid_api_token(server_url: str, user_id: int) -> typing.Optional[str]:
    """
    Return whether a user has a valid API token for a given Dataverse instance.

    :param server_url: the base URL of the Dataverse server
    :param user_id: the ID of an existing user
    :return: a valid API token stored for the user, or None
    :raise errors.UserDoesNotExistError: if not user with the given user ID
        exists
    :raise errors.DataverseNotReachableError: if there was an error during
        communication with the Dataverse API
    """
    api_token: typing.Optional[str] = settings.get_user_setting(user_id, 'DATAVERSE_API_TOKEN')
    if not api_token:
        # make sure the user even exists
        users.check_user_exists(user_id)
        return None
    if not is_api_token_valid(server_url, api_token):
        return None
    return api_token


def is_api_token_valid(server_url: str, api_token: str) -> bool:
    """
    Return whether an API token is valid for a given Dataverse instance.

    :param server_url: the base URL of the Dataverse server
    :param api_token: the API token to validate
    :return: whether or not the API token is valid
    :raise errors.DataverseNotReachableError: if there was an error during
        communication with the Dataverse API
    """
    # check basic API token structure
    if not all(c in '0123456789abcdef-' for c in api_token):
        return False
    parts = api_token.split('-')
    if len(parts) != 5:
        return False
    if [len(part) for part in parts] != [8, 4, 4, 4, 12]:
        return False
    # check whether the API token can actually be used
    try:
        r = requests.get(
            f'{server_url}/api/v1/users/token',
            headers={
                'X-Dataverse-key': api_token
            },
            allow_redirects=False,
            timeout=DATAVERSE_TIMEOUT
        )
    except requests.exceptions.RequestException:
        raise errors.DataverseNotReachableError()
    if r.status_code != 200:
        return False
    return True


def dataverse_has_required_metadata_blocks(server_url: str, api_token: str, dataverse: str) -> bool:
    """
    Return whether a Dataverse has the metadata blocks required for export.

    :param server_url: the base URL of a Dataverse server
    :param api_token: an API token for the Dataverse instance
    :param dataverse: the Dataverse to check
    :return: whether the Dataverse has the required metadata blocks
    :raise errors.DataverseNotReachableError: if there was an error during
        communication with the Dataverse API
    """
    try:
        r = requests.get(
            f'{server_url}/api/v1/dataverses/{dataverse}/metadatablocks',
            headers={
                'X-Dataverse-key': api_token
            },
            allow_redirects=False,
            timeout=DATAVERSE_TIMEOUT
        )
    except requests.exceptions.RequestException:
        raise errors.DataverseNotReachableError()
    if r.status_code != 200:
        # user might not even be allowed to access the Dataverse
        return False
    try:
        result = r.json()
        if result['status'] != 'OK':
            return False
        metadata_blocks = set()
        for metadata_block in result['data']:
            metadata_blocks.add(metadata_block['name'])
    except Exception:
        return False
    required_metadata_blocks = {'citation', 'process'}
    return all(
        metadata_block in metadata_blocks
        for metadata_block in required_metadata_blocks
    )



import enum
import datetime
import typing

from sqlalchemy.orm import Mapped, Query

from .. import db
from .utils import Model


@enum.unique
class BackgroundTaskStatus(enum.Enum):
    POSTED = 0
    CLAIMED = 1
    DONE = 2
    FAILED = 3

    def is_final(self) -> bool:
        return self in {
            BackgroundTaskStatus.DONE,
            BackgroundTaskStatus.FAILED
        }


class BackgroundTask(Model):
    __tablename__ = 'background_tasks'

    id: Mapped[int] = db.Column(db.Integer, primary_key=True)
    type: Mapped[str] = db.Column(db.Text, nullable=False)
    auto_delete: Mapped[bool] = db.Column(db.Boolean, nullable=False, default=True, server_default=db.true())
    data: Mapped[typing.Dict[str, typing.Any]] = db.Column(db.JSON, nullable=False)
    status: Mapped[BackgroundTaskStatus] = db.Column(db.Enum(BackgroundTaskStatus), nullable=False)
    result: Mapped[typing.Optional[typing.Dict[str, typing.Any]]] = db.Column(db.JSON, nullable=True)
    expiration_date: Mapped[typing.Optional[datetime.datetime]] = db.Column(db.TIMESTAMP(timezone=True), nullable=True)

    if typing.TYPE_CHECKING:
        query: typing.ClassVar[Query["BackgroundTask"]]

    @staticmethod
    def delete_expired_tasks() -> None:
        expired_tasks = BackgroundTask.query.filter(BackgroundTask.expiration_date <= datetime.datetime.now(datetime.timezone.utc)).all()
        for task in expired_tasks:
            db.session.delete(task)
        db.session.commit()

    def __repr__(self) -> str:
        return f'<{type(self).__name__}(id={self.id}, type={self.type}, auto_delete={self.auto_delete}, data={self.data}, status={self.status})>'



# coding: utf-8
"""
Drop description_as_html column from instruments table.
"""

import flask_sqlalchemy

from .utils import table_has_column


def run(db: flask_sqlalchemy.SQLAlchemy) -> bool:
    # Skip migration by condition
    if not table_has_column('instruments', 'description_as_html'):
        return False

    # Perform migration
    db.session.execute(db.text("""
            ALTER TABLE instruments
            DROP COLUMN description_as_html
        """))

    return True



# coding: utf-8
"""
Settings model.

To avoid creating a new table for each single type of user settings, the
Settings model allows storing all miscellaneous preferences for a user in
JSON format.
Some types of settings, e.g. notification settings, are complex enough to
justify custom models, while others are very simple, e.g. a single boolean
value or number, and can be stored using this model.
"""

import typing

from sqlalchemy.orm import Mapped, Query

from .. import db
from .users import User
from .utils import Model


class Settings(Model):
    __tablename__ = 'settings'

    user_id: Mapped[int] = db.Column(db.Integer, db.ForeignKey(User.id), primary_key=True)
    data: Mapped[typing.Dict[str, typing.Any]] = db.Column(db.JSON, nullable=False)

    if typing.TYPE_CHECKING:
        query: typing.ClassVar[Query["Settings"]]

    def __init__(
            self,
            user_id: int,
            data: typing.Dict[str, typing.Any]
    ) -> None:
        super().__init__(
            user_id=user_id,
            data=data
        )

    def __repr__(self) -> str:
        return f'<{type(self).__name__}(user_id={self.user_id}, data={self.data})>'



# coding: utf-8
"""

"""

import typing

from sqlalchemy.orm import Mapped, Query

from .. import db
from .utils import Model

__author__ = 'Du Kim Nguyen <k.nguyen@fz-juelich.de>'


class Language(Model):
    __tablename__ = 'languages'

    # default language IDs
    # offset to -100 to allow later addition of new default languages
    # see: migrations/languages_create_default_languages.py
    ENGLISH: int = -100 + 1
    GERMAN: int = -100 + 2

    id: Mapped[int] = db.Column(db.Integer, primary_key=True)
    lang_code: Mapped[str] = db.Column(db.String, nullable=False, unique=True)
    names: Mapped[typing.Dict[str, str]] = db.Column(db.JSON, nullable=False)
    datetime_format_datetime: Mapped[str] = db.Column(db.String, nullable=False)
    datetime_format_moment: Mapped[str] = db.Column(db.String, nullable=False)
    enabled_for_input: Mapped[bool] = db.Column(db.Boolean, nullable=False)
    enabled_for_user_interface: Mapped[bool] = db.Column(db.Boolean, nullable=False, default=False, server_default='FALSE')
    datetime_format_moment_output: Mapped[str] = db.Column(db.String, nullable=False, default='lll', server_default='lll')
    date_format_moment_output: Mapped[str] = db.Column(db.String, nullable=False, default='ll', server_default='ll')

    if typing.TYPE_CHECKING:
        query: typing.ClassVar[Query["Language"]]

    def __init__(
            self,
            names: typing.Dict[str, str],
            lang_code: str,
            datetime_format_datetime: str,
            datetime_format_moment: str,
            datetime_format_moment_output: str,
            date_format_moment_output: str,
            enabled_for_input: bool,
            enabled_for_user_interface: bool
    ) -> None:
        super().__init__(
            names=names,
            lang_code=lang_code,
            datetime_format_datetime=datetime_format_datetime,
            datetime_format_moment=datetime_format_moment,
            datetime_format_moment_output=datetime_format_moment_output,
            date_format_moment_output=date_format_moment_output,
            enabled_for_input=enabled_for_input,
            enabled_for_user_interface=enabled_for_user_interface
        )

    def __eq__(self, other: typing.Any) -> bool:
        if isinstance(other, Language):
            return bool(
                self.names == other.names and
                self.lang_code == other.lang_code and
                self.datetime_format_datetime == other.datetime_format_datetime and
                self.datetime_format_moment == other.datetime_format_moment and
                self.datetime_format_moment_output == other.datetime_format_moment_output and
                self.date_format_moment_output == other.date_format_moment_output and
                self.enabled_for_input == other.enabled_for_input and
                self.enabled_for_user_interface == other.enabled_for_user_interface
            )
        return NotImplemented

    def __repr__(self) -> str:
        return f'<{type(self).__name__}(id={self.id}, name={self.names.get("en", "Unknown")})>'



# coding: utf-8
"""

"""
import typing

from http import HTTPStatus
import flask
import flask_login
from flask_babel import _
from flask_wtf import FlaskForm
from wtforms import BooleanField, StringField
from wtforms import validators, ValidationError

from .. import frontend
from ...logic import users, errors, groups, projects, instruments
from ...logic.components import get_component
from ...logic.users import get_federated_identities, get_user_by_federated_user
from ..utils import validate_orcid
from ...utils import FlaskResponseT
from ...models import UserType


class UserReadOnlyForm(FlaskForm):
    action = StringField(validators=[validators.AnyOf(['toggle_read_only'])])
    should_be_read_only = BooleanField()


class UserHiddenForm(FlaskForm):
    action = StringField(validators=[validators.AnyOf(['toggle_hidden'])])
    should_be_hidden = BooleanField()


class UserActiveForm(FlaskForm):
    action = StringField(validators=[validators.AnyOf(['toggle_active'])])
    should_be_active = BooleanField()


class UserProfileForm(FlaskForm):
    action = StringField(validators=[validators.AnyOf(['edit_profile'])])
    orcid = StringField()
    affiliation = StringField()
    role = StringField()

    def validate_orcid(self, field: StringField) -> None:
        orcid = field.data
        # accept empty ORCID iDs
        if orcid is None:
            return
        orcid = orcid.strip()
        if not orcid:
            return
        # check ORCID iD syntax
        is_valid, orcid = validate_orcid(orcid)
        if not is_valid:
            raise ValidationError("Please enter a valid ORCID iD.")
        # keep sanitized ORCID iD on success
        field.data = orcid


@frontend.route('/users/me')
@flask_login.login_required
def current_user_profile() -> FlaskResponseT:
    return flask.redirect(flask.url_for('.user_profile', user_id=flask_login.current_user.id))


@frontend.route('/users/<int:user_id>', methods=['GET', 'POST'])
@flask_login.login_required
def user_profile(user_id: int) -> FlaskResponseT:
    try:
        user = users.get_user(user_id)
    except errors.UserDoesNotExistError:
        return flask.abort(404)

    last_modifying_user = None
    if user.last_modified_by_id is not None and user.last_modified_by_id != user.id:
        try:
            last_modifying_user = users.get_user(user.last_modified_by_id)
        except errors.UserDoesNotExistError:
            pass

    show_federated_identities = user.type not in (UserType.ELN_IMPORT_USER, UserType.FEDERATION_USER)
    fed_identity_descriptions_sampledb: dict[str, typing.Any] = {}
    fed_identity_descriptions_eln = []

    if show_federated_identities:
        for identity in sorted(get_federated_identities(user_id=user.id, component=None, active_status=True), key=lambda identity: identity.local_fed_id):
            if identity.local_fed_user.type == UserType.FEDERATION_USER:
                if identity.local_fed_user.component is None:
                    continue
                if identity.local_fed_user.component.uuid not in fed_identity_descriptions_sampledb:
                    fed_identity_descriptions_sampledb[identity.local_fed_user.component.uuid] = {
                        "component_name": identity.local_fed_user.component.name if identity.local_fed_user.component.name else _('Unknown database'),
                        "component_url": flask.url_for('.component', component_id=identity.local_fed_user.component_id),
                        "user_links": []
                    }
                component_address = identity.local_fed_user.component.address
                if component_address and not component_address.endswith("/"):
                    component_address += "/"
                fed_identity_descriptions_sampledb[identity.local_fed_user.component.uuid]["user_links"].append({
                    "user_url": f"{component_address}users/{identity.local_fed_user.fed_id}" if identity.local_fed_user.component.address else flask.url_for('.user_profile', user_id=identity.local_fed_id),
                    "fed_id": identity.local_fed_user.fed_id
                })
            elif identity.local_fed_user.type == UserType.ELN_IMPORT_USER:
                fed_identity_descriptions_eln.append({
                    "eln_object_id": identity.local_fed_user.eln_object_id if identity.local_fed_user.eln_object_id else None,
                    "eln_import_id": identity.local_fed_user.eln_import_id,
                    "eln_import_url": flask.url_for('.eln_import', eln_import_id=identity.local_fed_user.eln_import_id),
                    "user_id": identity.local_fed_id
                })

    for identity_description in fed_identity_descriptions_sampledb.values():
        identity_description["user_links"] = sorted(identity_description["user_links"], key=lambda x: x["fed_id"])

    show_local_federated_identity = user.type in [UserType.FEDERATION_USER, UserType.ELN_IMPORT_USER]
    local_federated_identity_user = get_user_by_federated_user(federated_user_id=user.id)

    if flask_login.current_user.is_admin and not flask_login.current_user.is_readonly:
        user_read_only_form = UserReadOnlyForm()
        user_read_only_form.should_be_read_only.default = not user.is_readonly
        if user_read_only_form.validate_on_submit():
            users.set_user_readonly(user.id, user_read_only_form.should_be_read_only.data)
            if user_read_only_form.should_be_read_only.data:
                flask.flash(_('The user has been marked as read only.'), 'success')
            else:
                flask.flash(_('The user has been unmarked as read only.'), 'success')
            return flask.redirect(flask.url_for('.user_profile', user_id=user.id))
        user_hidden_form = UserHiddenForm()
        user_hidden_form.should_be_hidden.default = not user.is_hidden
        if user_hidden_form.validate_on_submit():
            users.set_user_hidden(user.id, user_hidden_form.should_be_hidden.data)
            if user_hidden_form.should_be_hidden.data:
                flask.flash(_('The user has been marked as hidden.'), 'success')
            else:
                flask.flash(_('The user has been unmarked as hidden.'), 'success')
            return flask.redirect(flask.url_for('.user_profile', user_id=user.id))
        user_active_form = UserActiveForm()
        user_active_form.should_be_active.default = not user.is_active
        if user_active_form.validate_on_submit():
            users.set_user_active(user.id, user_active_form.should_be_active.data)
            if user_active_form.should_be_active.data:
                flask.flash(_('The user has been activated.'), 'success')
            else:
                flask.flash(_('The user has been deactivated.'), 'success')
            return flask.redirect(flask.url_for('.user_profile', user_id=user.id))
        if user.component_id is None:
            user_profile_form = UserProfileForm()
            if not user_profile_form.is_submitted():
                user_profile_form.orcid.data = user.orcid
                user_profile_form.role.data = user.role
                user_profile_form.affiliation.data = user.affiliation
            if user_profile_form.validate_on_submit():
                if user_profile_form.orcid.data and user_profile_form.orcid.data.strip():
                    orcid = user_profile_form.orcid.data.strip()
                else:
                    orcid = None
                if user_profile_form.affiliation.data and user_profile_form.affiliation.data.strip():
                    affiliation = user_profile_form.affiliation.data.strip()
                else:
                    affiliation = None
                if user_profile_form.role.data and user_profile_form.role.data.strip():
                    role = user_profile_form.role.data.strip()
                else:
                    role = None
                extra_fields = {}
                for extra_field_id in flask.current_app.config['EXTRA_USER_FIELDS']:
                    extra_field_value = flask.request.form.get('extra_field_' + str(extra_field_id))
                    if extra_field_value and extra_field_value.strip():
                        extra_fields[extra_field_id] = extra_field_value
                change_orcid = (user.orcid != orcid and (orcid is not None or user.orcid is not None))
                change_affiliation = (user.affiliation != affiliation and (affiliation is not None or user.affiliation is not None))
                change_role = (user.role != role and (role is not None or user.role is not None))
                change_extra_fields = user.extra_fields != extra_fields
                if change_orcid or change_affiliation or change_role or change_extra_fields:
                    users.update_user(
                        user.id,
                        updating_user_id=flask_login.current_user.id,
                        orcid=orcid,
                        affiliation=affiliation,
                        role=role,
                        extra_fields=extra_fields
                    )
                    flask.flash(_("Successfully updated the user information."), 'success')
                    return flask.redirect(flask.url_for('.user_profile', user_id=user.id))
        else:
            user_profile_form = None
    elif flask.request.method.lower() == 'post':
        return flask.abort(HTTPStatus.METHOD_NOT_ALLOWED)
    else:
        user_read_only_form = None
        user_hidden_form = None
        user_active_form = None
        user_profile_form = None
    return flask.render_template(
        'profile.html',
        user_read_only_form=user_read_only_form,
        user_hidden_form=user_hidden_form,
        user_active_form=user_active_form,
        user_profile_form=user_profile_form,
        user=user,
        last_modifying_user=last_modifying_user,
        EXTRA_USER_FIELDS=flask.current_app.config['EXTRA_USER_FIELDS'],
        basic_groups=groups.get_user_groups(user.id),
        project_groups=projects.get_user_projects(user.id, include_groups=True),
        instruments=[
            instruments.get_instrument(instrument_id)
            for instrument_id in instruments.get_user_instruments(user.id, exclude_hidden=True)
        ],
        get_component=get_component,
        show_federated_identities=show_federated_identities,
        fed_identity_descriptions_sampledb=fed_identity_descriptions_sampledb,
        fed_identity_descriptions_eln=fed_identity_descriptions_eln,
        local_federated_identity_user=local_federated_identity_user,
        show_local_federated_identity=show_local_federated_identity
    )



import typing
import datetime

from . import errors
from .. import db
from ..models import UpdatableObjectsCheck, UpdatableObjectsCheckStatus, AutomaticSchemaUpdate, AutomaticSchemaUpdateStatus
from . import background_tasks


def get_updatable_objects_checks() -> typing.Sequence[UpdatableObjectsCheck]:
    return UpdatableObjectsCheck.query.order_by(db.desc(UpdatableObjectsCheck.utc_datetime)).all()


def start_updatable_objects_checks(
        user_id: int,
        action_ids: typing.Optional[typing.Sequence[int]]
) -> UpdatableObjectsCheck:
    if db.session.query(db.exists().where(UpdatableObjectsCheck.status != UpdatableObjectsCheckStatus.DONE)).scalar():
        raise errors.UpdatableObjectsCheckAlreadyInProgressError()
    updatable_objects_check = UpdatableObjectsCheck(
        status=UpdatableObjectsCheckStatus.POSTED,
        user_id=user_id,
        utc_datetime=datetime.datetime.now(tz=datetime.timezone.utc).replace(tzinfo=None),
        action_ids=action_ids,
        result=None,
    )
    db.session.add(updatable_objects_check)
    db.session.commit()
    background_tasks.post_check_for_automatic_schema_updates_task(updatable_objects_check_id=updatable_objects_check.id)
    return updatable_objects_check


def get_updatable_objects_check(
        updatable_objects_check_id: int
) -> UpdatableObjectsCheck:
    updatable_objects_check = UpdatableObjectsCheck.query.filter_by(id=updatable_objects_check_id).first()
    if updatable_objects_check is None:
        raise errors.UpdatableObjectsCheckDoesNotExistError()
    return updatable_objects_check


def set_updatable_objects_check_status(
        updatable_objects_check_id: int,
        status: UpdatableObjectsCheckStatus,
        result: typing.Dict[str, typing.Any]
) -> None:
    updatable_objects_check = UpdatableObjectsCheck.query.filter_by(id=updatable_objects_check_id).first()
    if updatable_objects_check is None:
        raise errors.UpdatableObjectsCheckDoesNotExistError()
    updatable_objects_check.status = status
    updatable_objects_check.result = result
    db.session.add(updatable_objects_check)
    db.session.commit()


def get_automatic_schema_updates() -> typing.Sequence[AutomaticSchemaUpdate]:
    return AutomaticSchemaUpdate.query.order_by(db.desc(AutomaticSchemaUpdate.utc_datetime)).all()


def start_automatic_schema_updates(
        user_id: int,
        updatable_objects_check_id: int,
        object_ids: typing.Sequence[int],
) -> AutomaticSchemaUpdate:
    if db.session.query(db.exists().where(AutomaticSchemaUpdate.status != AutomaticSchemaUpdateStatus.DONE)).scalar():
        raise errors.AutomaticSchemaUpdateAlreadyInProgressError()
    automatic_schema_update = AutomaticSchemaUpdate(
        status=AutomaticSchemaUpdateStatus.POSTED,
        user_id=user_id,
        utc_datetime=datetime.datetime.now(tz=datetime.timezone.utc).replace(tzinfo=None),
        updatable_objects_check_id=updatable_objects_check_id,
        object_ids=object_ids,
        result=None,
    )
    db.session.add(automatic_schema_update)
    db.session.commit()
    background_tasks.post_perform_automatic_schema_updates_task(automatic_schema_update_id=automatic_schema_update.id)
    return automatic_schema_update


def get_automatic_schema_update(
        automatic_schema_update_id: int
) -> AutomaticSchemaUpdate:
    automatic_schema_update = AutomaticSchemaUpdate.query.filter_by(id=automatic_schema_update_id).first()
    if automatic_schema_update is None:
        raise errors.AutomaticSchemaUpdateDoesNotExistError()
    return automatic_schema_update


def set_automatic_schema_update_status(
        automatic_schema_update_id: int,
        status: AutomaticSchemaUpdateStatus,
        result: typing.Dict[str, typing.Any]
) -> None:
    automatic_schema_update = AutomaticSchemaUpdate.query.filter_by(id=automatic_schema_update_id).first()
    if automatic_schema_update is None:
        raise errors.AutomaticSchemaUpdateDoesNotExistError()
    automatic_schema_update.status = status
    automatic_schema_update.result = result
    db.session.add(automatic_schema_update)
    db.session.commit()



# coding: utf-8
"""
Create default languages.
"""

import json

import flask_sqlalchemy

from .utils import table_has_column
from ..languages import Language


def run(db: flask_sqlalchemy.SQLAlchemy) -> bool:
    default_languages = [
        {
            'id': Language.ENGLISH,
            'names': json.dumps({
                'en': 'English',
                'de': 'Englisch'
            }),
            'lang_code': 'en',
            'datetime_format_datetime': '%Y-%m-%d %H:%M:%S',
            'datetime_format_moment': 'YYYY-MM-DD HH:mm:ss',
            'enabled_for_input': True,
        },
        {
            'id': Language.GERMAN,
            'names': json.dumps({
                'en': 'German',
                'de': 'Deutsch'
            }),
            'lang_code': 'de',
            'datetime_format_datetime': '%d.%m.%Y %H:%M:%S',
            'datetime_format_moment': 'DD.MM.YYYY HH:mm:ss',
            'enabled_for_input': False,
        }
    ]

    existing_language_ids = [
        language[0]
        for language in db.session.execute(db.text("""
               SELECT id
               FROM languages;
           """)).fetchall()
    ]

    performed_migration = False
    for language in default_languages:
        # Skip migration by condition
        if language['id'] in existing_language_ids:
            continue

        # Perform migration
        db.session.execute(db.text("""
               INSERT INTO languages (id, names, lang_code, datetime_format_datetime, datetime_format_moment, enabled_for_input)
               VALUES (:id, :names, :lang_code, :datetime_format_datetime, :datetime_format_moment, :enabled_for_input);
           """), params=language)
        performed_migration = True

    if table_has_column('languages', 'enabled_for_user_interface'):
        db.session.execute(db.text("""
            UPDATE languages
            SET enabled_for_user_interface = TRUE
            WHERE id < 0
        """))
    if table_has_column('languages', 'datetime_format_moment_output'):
        db.session.execute(db.text("""
            UPDATE languages
            SET datetime_format_moment_output = 'MMM D, YYYY, h:mm:ss A'
            WHERE id = -99
        """))
        db.session.execute(db.text("""
            UPDATE languages
            SET datetime_format_moment_output = 'DD.MM.YYYY HH:mm:ss'
            WHERE id = -98
        """))
    if table_has_column('languages', 'date_format_moment_output'):
        db.session.execute(db.text("""
            UPDATE languages
            SET date_format_moment_output = 'MMM D, YYYY'
            WHERE id = -99
        """))
        db.session.execute(db.text("""
            UPDATE languages
            SET date_format_moment_output = 'DD.MM.YYYY'
            WHERE id = -98
        """))

    return performed_migration



# coding: utf-8
"""
Add missing NOT NULL constraints to foreign key columns in action_translations table.
"""

import flask_sqlalchemy

from .utils import column_is_nullable


def run(db: flask_sqlalchemy.SQLAlchemy) -> bool:
    # Skip migration by condition
    if not column_is_nullable('action_translations', 'action_id') and not column_is_nullable('action_translations', 'language_id'):
        return False

    # Perform migration
    db.session.execute(db.text("""
        ALTER TABLE action_translations
            ALTER COLUMN action_id SET NOT NULL,
            ALTER COLUMN language_id SET NOT NULL
    """))
    return True



# coding: utf-8
"""
Implementation of validate(instance, schema)
"""

import re
import datetime
import decimal
import string
import typing
import math
import json

import plotly
from flask_babel import _
import flask

from .conditions import are_conditions_fulfilled
from ...logic import actions, objects, datatypes, users, languages
from ...models import ActionType
from ..errors import ObjectDoesNotExistError, ValidationError, ValidationMultiError, UserDoesNotExistError, InvalidURLError
from .utils import units_are_valid
from ..utils import get_translated_text, parse_url
from ..units import get_dimensionality_for_units, get_magnitude_in_base_units, get_old_dimensionality

OPT_IMPORT_KEYS = {'export_edit_note', 'component_uuid', 'eln_source_url', 'eln_object_url', 'eln_user_url'}


def validate(
        instance: typing.Union[typing.Dict[str, typing.Any], typing.List[typing.Any]],
        schema: typing.Dict[str, typing.Any],
        path: typing.Optional[typing.List[str]] = None,
        allow_disabled_languages: bool = False,
        strict: bool = False,
        file_names_by_id: typing.Optional[typing.Dict[int, str]] = None
) -> None:
    """
    Validates the given instance using the given schema and raises a ValidationError if it is invalid.

    :param instance: the sampledb object
    :param schema: the valid sampledb object schema
    :param path: the path to this subinstance / subschema
    :param allow_disabled_languages: whether disabled languages are allowed
    :param strict: whether the data should be evaluated in strict mode, or backwards compatible otherwise
    :param file_names_by_id: a dict mapping file IDs to file names, or None
    :raise ValidationError: if the schema is invalid.
    """
    if path is None:
        path = []
    if not isinstance(schema, dict):
        raise ValidationError('invalid schema (must be dict)', path)
    if 'type' not in schema:
        raise ValidationError('invalid schema (must contain type)', path)
    if schema['type'] == 'array' and isinstance(instance, list):
        return _validate_array(instance, schema, path, allow_disabled_languages=allow_disabled_languages, strict=strict, file_names_by_id=file_names_by_id)
    elif schema['type'] == 'object' and isinstance(instance, dict):
        return _validate_object(instance, schema, path, allow_disabled_languages=allow_disabled_languages, strict=strict, file_names_by_id=file_names_by_id)
    elif schema['type'] == 'text' and isinstance(instance, dict):
        return _validate_text(instance, schema, path, allow_disabled_languages=allow_disabled_languages)
    elif schema['type'] == 'datetime' and isinstance(instance, dict):
        return _validate_datetime(instance, schema, path)
    elif schema['type'] == 'bool' and isinstance(instance, dict):
        return _validate_bool(instance, schema, path)
    elif schema['type'] == 'quantity' and isinstance(instance, dict):
        return _validate_quantity(instance, schema, path)
    elif schema['type'] == 'sample' and isinstance(instance, dict):
        return _validate_sample(instance, schema, path)
    elif schema['type'] == 'measurement' and isinstance(instance, dict):
        return _validate_measurement(instance, schema, path)
    elif schema['type'] == 'object_reference' and isinstance(instance, dict):
        return _validate_object_reference(instance, schema, path)
    elif schema['type'] == 'tags' and isinstance(instance, dict):
        return _validate_tags(instance, schema, path, strict=strict)
    elif schema['type'] == 'hazards' and isinstance(instance, dict):
        return _validate_hazards(instance, schema, path)
    elif schema['type'] == 'user' and isinstance(instance, dict):
        return _validate_user(instance, schema, path)
    elif schema['type'] == 'plotly_chart' and isinstance(instance, dict):
        return _validate_plotly_chart(instance, schema, path)
    elif schema['type'] == 'timeseries' and isinstance(instance, dict):
        return _validate_timeseries(instance, schema, path)
    elif schema['type'] == 'file' and isinstance(instance, dict):
        return _validate_file(instance, schema, path, file_names_by_id=file_names_by_id)
    else:
        raise ValidationError('invalid type', path)


def _validate_array(
        instance: typing.List[typing.Any],
        schema: typing.Dict[str, typing.Any], path: typing.List[str],
        allow_disabled_languages: bool = False,
        strict: bool = False,
        file_names_by_id: typing.Optional[typing.Dict[int, str]] = None
) -> None:
    """
    Validates the given instance using the given array schema and raises a ValidationError if it is invalid.

    :param instance: the sampledb object
    :param schema: the valid sampledb object schema
    :param path: the path to this subinstance / subschema
    :param strict: whether the data should be evaluated in strict mode, or backwards compatible otherwise
    :param file_names_by_id: a dict mapping file IDs to file names, or None
    :raise ValidationError: if the schema is invalid.
    """
    if not isinstance(instance, list):
        raise ValidationError('instance must be list', path)
    if 'minItems' in schema and len(instance) < schema['minItems']:
        raise ValidationError(f'expected at least {schema["minItems"]} items', path)
    if 'maxItems' in schema and len(instance) > schema['maxItems']:
        raise ValidationError(f'expected at most {schema["maxItems"]} items', path)
    errors = []
    for index, item in enumerate(instance):
        try:
            validate(item, schema['items'], path + [str(index)], allow_disabled_languages=allow_disabled_languages, strict=strict, file_names_by_id=file_names_by_id)
        except ValidationError as e:
            errors.append(e)
    if len(errors) == 1:
        raise errors[0]
    elif len(errors) > 1:
        raise ValidationMultiError(errors)


def _validate_hazards(instance: typing.Dict[str, typing.Any], schema: typing.Dict[str, typing.Any], path: typing.List[str]) -> None:
    """
    Validate the given instance using the given GHS hazards schema and raise a ValidationError if it is invalid.

    :param instance: the sampledb object
    :param schema: the valid sampledb object schema
    :param path: the path to this subinstance / subschema
    :raise ValidationError: if the schema is invalid.
    """
    if not isinstance(instance, dict):
        raise ValidationError('instance must be dict', path)
    if path != ['hazards']:
        raise ValidationError('GHS hazards must be a top-level entry named "hazards"', path)
    valid_keys = {'_type', 'hazards'}
    required_keys = valid_keys
    schema_keys = set(instance.keys())
    invalid_keys = schema_keys - valid_keys - OPT_IMPORT_KEYS
    if invalid_keys:
        raise ValidationError(f'unexpected keys in schema: {invalid_keys}', path)
    missing_keys = required_keys - schema_keys
    if missing_keys:
        raise ValidationError(f'missing keys in schema: {missing_keys}', path)
    if instance['_type'] != 'hazards':
        raise ValidationError('expected _type "hazards"', path)
    if not isinstance(instance['hazards'], list):
        raise ValidationError('hazards must be list', path)
    errors = []
    hazards = []
    for index, item in enumerate(instance['hazards']):
        if not isinstance(item, int):
            errors.append(ValidationError(f'invalid hazard index type: {type(item)}', path + ['hazards', str(index)]))
        elif item in hazards:
            errors.append(ValidationError(f'duplicate hazard index: {item}', path + ['hazards', str(index)]))
        elif item < 1 or item > 9:
            errors.append(ValidationError(f'invalid hazard index: {item}', path + ['hazards', str(index)]))
        else:
            hazards.append(item)

    if len(errors) == 1:
        raise errors[0]
    elif len(errors) > 1:
        raise ValidationMultiError(errors)


def _validate_tags(
        instance: typing.Dict[str, typing.Any],
        schema: typing.Dict[str, typing.Any], path: typing.List[str],
        strict: bool = False
) -> None:
    """
    Validates the given instance using the given tags schema and raises a ValidationError if it is invalid.

    :param instance: the sampledb object
    :param schema: the valid sampledb object schema
    :param path: the path to this subinstance / subschema
    :param strict: whether the data should be evaluated in strict mode, or backwards compatible otherwise
    :raise ValidationError: if the schema is invalid.
    """
    if not isinstance(instance, dict):
        raise ValidationError('instance must be dict', path)
    valid_keys = {'_type', 'tags'}
    required_keys = valid_keys
    schema_keys = set(instance.keys())
    invalid_keys = schema_keys - valid_keys - OPT_IMPORT_KEYS
    if invalid_keys:
        raise ValidationError(f'unexpected keys in schema: {invalid_keys}', path)
    missing_keys = required_keys - schema_keys
    if missing_keys:
        raise ValidationError(f'missing keys in schema: {missing_keys}', path)
    if instance['_type'] != 'tags':
        raise ValidationError('expected _type "tags"', path)
    if not isinstance(instance['tags'], list):
        raise ValidationError('tags must be list', path)
    errors = []
    tags = []
    for index, item in enumerate(instance['tags']):
        if not isinstance(item, str):
            errors.append(ValidationError(f'invalid tag type: {type(item)}', path + ['tags', str(index)]))
        elif item in tags:
            errors.append(ValidationError(f'duplicate tag: {item}', path + ['tags', str(index)]))
        elif item.lower() != item:
            errors.append(ValidationError(f'tag not lowercase: {item}', path + ['tags', str(index)]))
        elif any(c not in 'abcdefghijklmnopqrstuvwxyz0123456789_-äöüß' for c in item):
            errors.append(ValidationError(f'tag contains invalid character: {item}', path + ['tags', str(index)]))
        elif strict and all(c in string.digits for c in item) and not flask.current_app.config['ENABLE_NUMERIC_TAGS']:
            errors.append(ValidationError('numeric tags are not supported', path + ['tags', str(index)]))
        else:
            tags.append(item)

    if len(errors) == 1:
        raise errors[0]
    elif len(errors) > 1:
        raise ValidationMultiError(errors)


def _validate_object(
        instance: typing.Dict[str, typing.Any],
        schema: typing.Dict[str, typing.Any],
        path: typing.List[str],
        allow_disabled_languages: bool = False,
        strict: bool = False,
        file_names_by_id: typing.Optional[typing.Dict[int, str]] = None
) -> None:
    """
    Validates the given instance using the given object schema and raises a ValidationError if it is invalid.

    :param instance: the sampledb object
    :param schema: the valid sampledb object schema
    :param path: the path to this subinstance / subschema
    :param strict: whether the data should be evaluated in strict mode, or backwards compatible otherwise
    :param file_names_by_id: a dict mapping file IDs to file names, or None
    :raise ValidationError: if the schema is invalid.
    """
    if not isinstance(instance, dict):
        raise ValidationError('instance must be dict', path)
    errors = []

    properties_with_unfulfilled_conditions = []
    for property_name, property_schema in schema['properties'].items():
        if not are_conditions_fulfilled(property_schema.get('conditions'), instance):
            properties_with_unfulfilled_conditions.append(property_name)
            if property_name in instance or (property_name == 'name' and not path):
                errors.append(ValidationError(f'conditions for property "{property_name}" not fulfilled', path + [property_name]))

    if 'required' in schema:
        for property_name in schema['required']:
            if property_name in properties_with_unfulfilled_conditions:
                # this property must not be included, as its conditions are not fulfilled
                continue
            if property_name not in instance:
                errors.append(ValidationError(f'missing required property "{property_name}"', path + [property_name]))
    for property_name, property_value in instance.items():
        try:
            if property_name not in schema['properties']:
                raise ValidationError(f'unknown property "{property_name}"', path + [property_name])
            else:
                validate(property_value, schema['properties'][property_name], path + [property_name], allow_disabled_languages=allow_disabled_languages, strict=strict, file_names_by_id=file_names_by_id)
        except ValidationError as e:
            errors.append(e)
    if len(errors) == 1:
        raise errors[0]
    elif len(errors) > 1:
        raise ValidationMultiError(errors)


def _validate_text(instance: typing.Dict[str, typing.Any], schema: typing.Dict[str, typing.Any], path: typing.List[str], allow_disabled_languages: bool = False) -> None:
    """
    Validates the given instance using the given text object schema and raises a ValidationError if it is invalid.

    :param instance: the sampledb object
    :param schema: the valid sampledb object schema
    :param path: the path to this subinstance / subschema
    :raise ValidationError: if the schema is invalid.
    """
    if not isinstance(instance, dict):
        raise ValidationError('instance must be dict', path)
    valid_keys = {'_type', 'text'}
    required_keys = set(valid_keys)
    if schema.get('markdown', False):
        valid_keys.add('is_markdown')
    schema_keys = set(instance.keys())
    invalid_keys = schema_keys - valid_keys - OPT_IMPORT_KEYS
    if invalid_keys:
        raise ValidationError(f'unexpected keys in schema: {invalid_keys}', path)
    missing_keys = required_keys - schema_keys
    if missing_keys:
        raise ValidationError(f'missing keys in schema: {missing_keys}', path)
    if instance['_type'] != 'text':
        raise ValidationError('expected _type "text"', path)
    if not isinstance(instance['text'], str) and not isinstance(instance['text'], dict):
        raise ValidationError('text must be str or a dictionary', path)
    choices = schema.get('choices', None)
    if choices:
        for choice in choices:
            if str(choice) == instance['text'] or choice == {'en': instance['text']} or instance['text'] == {'en': choice}:
                instance['text'] = choice
                break
        if instance['text'] not in choices:
            raise ValidationError(_('The text must be one of: %(choices)s.', choices=', '.join(get_translated_text(choice) for choice in choices)), path)
    min_length = schema.get('minLength', 0)
    max_length = schema.get('maxLength', None)

    if isinstance(instance['text'], str):
        if len(instance['text']) < min_length:
            raise ValidationError(_('The text must be at least %(min_length)s characters long.', min_length=min_length), path)
        if max_length is not None and len(instance['text']) > max_length:
            raise ValidationError(_('The text must be at most %(max_length)s characters long.', max_length=max_length), path)
    elif not choices:
        all_languages = languages.get_languages()
        language_names = {
            language.lang_code: get_translated_text(language.names)
            for language in all_languages
        }
        if 'languages' in schema:
            if schema['languages'] == 'all':
                allowed_language_codes = {
                    language.lang_code
                    for language in all_languages
                    if allow_disabled_languages or language.enabled_for_input
                }
            else:
                allowed_language_codes = {
                    language.lang_code
                    for language in all_languages
                    if allow_disabled_languages or (language.enabled_for_input and language.lang_code in schema['languages'])
                }
        else:
            allowed_language_codes = {'en'}
        for text in instance['text'].values():
            if not isinstance(text, str):
                raise ValidationError('The text in a translation dict must be str.', path)
            if len(text) < min_length:
                raise ValidationError(_('The text must be at least %(min_length)s characters long.', min_length=min_length), path)
            if max_length is not None and len(text) > max_length:
                raise ValidationError(_('The text must be at most %(max_length)s characters long.', max_length=max_length), path)
        for lang in instance['text'].keys():
            if lang not in allowed_language_codes:
                raise ValidationError(_('The language "%(lang_code)s" is not allowed for this field.', lang_code=language_names.get(lang, lang)), path)
    if 'pattern' in schema:
        if not isinstance(schema['pattern'], str):
            raise ValidationError('pattern must be str', path)
        if isinstance(instance['text'], dict):
            for text in instance['text'].values():
                if re.match(schema['pattern'], text) is None:
                    raise ValidationError(_('Input must match: %(pattern)s', pattern=schema['pattern']), path)
        if isinstance(instance['text'], str):
            if re.match(schema['pattern'], instance['text']) is None:
                raise ValidationError(_('Input must match: %(pattern)s', pattern=schema['pattern']), path)
    if 'is_markdown' in instance and not isinstance(instance['is_markdown'], bool):
        raise ValidationError('is_markdown must be bool', path)


def _validate_datetime(instance: typing.Dict[str, typing.Any], schema: typing.Dict[str, typing.Any], path: typing.List[str]) -> None:
    """
    Validates the given instance using the given datetime object schema and raises a ValidationError if it is invalid.

    :param instance: the sampledb object
    :param schema: the valid sampledb object schema
    :param path: the path to this subinstance / subschema
    :raise ValidationError: if the schema is invalid.
    """
    if not isinstance(instance, dict):
        raise ValidationError('instance must be dict', path)
    valid_keys = {'_type', 'utc_datetime'}
    required_keys = valid_keys
    schema_keys = set(instance.keys())
    invalid_keys = schema_keys - valid_keys - OPT_IMPORT_KEYS
    if invalid_keys:
        raise ValidationError(f'unexpected keys in schema: {invalid_keys}', path)
    missing_keys = required_keys - schema_keys
    if missing_keys:
        raise ValidationError(f'missing keys in schema: {missing_keys}', path)
    if instance['_type'] != 'datetime':
        raise ValidationError('expected _type "datetime"', path)
    if not isinstance(instance['utc_datetime'], str):
        raise ValidationError('utc_datetime must be str', path)
    try:
        datetime.datetime.strptime(instance['utc_datetime'], '%Y-%m-%d %H:%M:%S')
    except ValueError:
        raise ValidationError(_('Please enter the date and time in the format: %(datetime_format)s', datetime_format='YYYY-MM-DD HH:MM:SS'), path)


def _validate_bool(instance: typing.Dict[str, typing.Any], schema: typing.Dict[str, typing.Any], path: typing.List[str]) -> None:
    """
    Validates the given instance using the given boolean object schema and raises a ValidationError if it is invalid.

    :param instance: the sampledb object
    :param schema: the valid sampledb object schema
    :param path: the path to this subinstance / subschema
    :raise ValidationError: if the schema is invalid.
    """
    if not isinstance(instance, dict):
        raise ValidationError('instance must be dict', path)
    valid_keys = {'_type', 'value'}
    required_keys = valid_keys
    schema_keys = set(instance.keys())
    invalid_keys = schema_keys - valid_keys - OPT_IMPORT_KEYS
    if invalid_keys:
        raise ValidationError(f'unexpected keys in schema: {invalid_keys}', path)
    missing_keys = required_keys - schema_keys
    if missing_keys:
        raise ValidationError(f'missing keys in schema: {missing_keys}', path)
    if instance['_type'] != 'bool':
        raise ValidationError('expected _type "bool"', path)
    if not isinstance(instance['value'], bool):
        raise ValidationError('value must be bool', path)


def _validate_quantity(instance: typing.Dict[str, typing.Any], schema: typing.Dict[str, typing.Any], path: typing.List[str]) -> None:
    """
    Validates the given instance using the given quantity object schema and raises a ValidationError if it is invalid.

    :param instance: the sampledb object
    :param schema: the valid sampledb object schema
    :param path: the path to this subinstance / subschema
    :raise ValidationError: if the schema is invalid.
    """
    if not isinstance(instance, dict):
        raise ValidationError('instance must be dict', path)
    required_keys = {'_type'}
    valid_keys = required_keys.union({'units', 'dimensionality', 'magnitude_in_base_units', 'magnitude'})
    schema_keys = set(instance.keys())
    invalid_keys = schema_keys - valid_keys - OPT_IMPORT_KEYS
    if invalid_keys:
        raise ValidationError(f'unexpected keys in schema: {invalid_keys}', path)
    missing_keys = required_keys - schema_keys
    if missing_keys:
        raise ValidationError(f'missing keys in schema: {missing_keys}', path)
    if instance['_type'] != 'quantity':
        raise ValidationError('expected _type "quantity"', path)
    if 'units' not in instance:
        if not isinstance(schema['units'], str):
            raise ValidationError('missing unambiguous units for quantity property with multiple possible units', path)
        instance['units'] = schema['units']
    if not isinstance(instance['units'], str):
        raise ValidationError('units must be str', path)
    if not units_are_valid(instance['units']):
        raise ValidationError('Invalid/Unknown units', path)
    try:
        if isinstance(schema['units'], str):
            schema_units = schema['units']
        else:
            schema_units = schema['units'][0]
        schema_quantity = datatypes.Quantity(1.0, units=schema_units)
    except Exception:
        raise ValidationError('Unable to create schema quantity', path)

    quantity_magnitude = None
    quantity_magnitude_in_base_units = None

    if 'magnitude' in instance:
        if not isinstance(instance['magnitude'], float) and not isinstance(instance['magnitude'], int):
            raise ValidationError('magnitude must be float or int', path)
        if not math.isfinite(instance['magnitude']):
            raise ValidationError('magnitude must be a finite number', path)
        try:
            quantity_magnitude = datatypes.Quantity(instance['magnitude'], units=instance['units'],
                                                    already_in_base_units=False)
        except Exception:
            raise ValidationError('Unable to create quantity based on given magnitude', path)
    if 'magnitude_in_base_units' in instance:
        if not isinstance(instance['magnitude_in_base_units'], float) and not isinstance(instance['magnitude_in_base_units'], int):
            raise ValidationError('magnitude_in_base_units must be float or int', path)
        if not math.isfinite(instance['magnitude_in_base_units']):
            raise ValidationError('magnitude_in_base_units must be a finite number', path)
        try:
            quantity_magnitude_in_base_units = datatypes.Quantity(instance['magnitude_in_base_units'], units=instance['units'], already_in_base_units=True)
        except Exception:
            raise ValidationError('Unable to create quantity based on given magnitude_in_base_units', path)

    if quantity_magnitude is not None and quantity_magnitude_in_base_units is not None and not (
            math.isclose(quantity_magnitude.magnitude, quantity_magnitude_in_base_units.magnitude) or
            math.isclose(quantity_magnitude.magnitude_in_base_units, quantity_magnitude_in_base_units.magnitude_in_base_units)
    ):
        raise ValidationError('magnitude and magnitude_in_base_units do not match, either set only one or make sure both match', path)
    if quantity_magnitude is None:
        if quantity_magnitude_in_base_units is None:
            raise ValidationError('missing keys in schema: either magnitude or magnitude_in_base_units has to be given', path)
        quantity_magnitude = quantity_magnitude_in_base_units

    if 'min_magnitude' in schema and quantity_magnitude.magnitude_in_base_units < schema['min_magnitude']:
        min_magnitude = datatypes.Quantity(schema["min_magnitude"], units=schema['units'], already_in_base_units=True).magnitude
        min_value = f'{min_magnitude}{" " + schema["units"] if schema["units"] != "1" else ""}'
        raise ValidationError(_('Must be greater than or equal to %(value)s', value=min_value), path)

    if 'max_magnitude' in schema and quantity_magnitude.magnitude_in_base_units > schema['max_magnitude']:
        max_magnitude = datatypes.Quantity(schema["max_magnitude"], units=schema['units'], already_in_base_units=True).magnitude
        max_value = f'{max_magnitude}{" " + schema["units"] if schema["units"] != "1" else ""}'
        raise ValidationError(_('Must be less than or equal to %(value)s', value=max_value), path)

    # automatically add dimensionality and either magnitude oder magnitude_in_base_units, if they haven't been given yet
    for key, value in quantity_magnitude.to_json().items():
        if key not in instance:
            instance[key] = value

    if not isinstance(instance['dimensionality'], str):
        raise ValidationError('dimensionality must be str', path)
    if quantity_magnitude.dimensionality != schema_quantity.dimensionality:
        raise ValidationError(f'Invalid units, expected units for dimensionality "{str(schema_quantity.dimensionality)}"', path)
    if str(quantity_magnitude.dimensionality) != instance['dimensionality'] and get_old_dimensionality(quantity_magnitude.units) != instance['dimensionality']:
        raise ValidationError(f'Invalid dimensionality, expected "{str(schema_quantity.dimensionality)}"', path)


def _validate_sample(instance: typing.Dict[str, typing.Any], schema: typing.Dict[str, typing.Any], path: typing.List[str]) -> None:
    """
    Validates the given instance using the given sample object schema and raises a ValidationError if it is invalid.

    :param instance: the sampledb object
    :param schema: the valid sampledb object schema
    :param path: the path to this subinstance / subschema
    :raise ValidationError: if the schema is invalid.
    """
    if not isinstance(instance, dict):
        raise ValidationError('instance must be dict', path)
    valid_keys = {'_type', 'object_id'}
    required_keys = valid_keys
    schema_keys = set(instance.keys())
    invalid_keys = schema_keys - valid_keys - OPT_IMPORT_KEYS
    if invalid_keys:
        raise ValidationError(f'unexpected keys in schema: {invalid_keys}', path)
    missing_keys = required_keys - schema_keys
    if missing_keys:
        raise ValidationError(f'missing keys in schema: {missing_keys}', path)
    if instance['_type'] != 'sample':
        raise ValidationError('expected _type "sample"', path)
    if not isinstance(instance['object_id'], int):
        raise ValidationError('object_id must be int', path)
    if 'component_uuid' in instance and instance['component_uuid'] != flask.current_app.config['FEDERATION_UUID']:
        pass
    elif 'eln_source_url' in instance:
        validate_eln_urls(instance, path)
    else:
        try:
            sample = objects.get_object(object_id=instance['object_id'])
        except ObjectDoesNotExistError:
            raise ValidationError('object does not exist', path)
        if sample.action_id is None:
            raise ValidationError('object must be sample', path)
        action = actions.get_action(sample.action_id)
        if action.type is None:
            raise ValidationError('object must be sample', path)
        if ActionType.SAMPLE_CREATION not in {action.type_id, action.type.fed_id}:
            raise ValidationError('object must be sample', path)


def _validate_measurement(instance: typing.Dict[str, typing.Any], schema: typing.Dict[str, typing.Any], path: typing.List[str]) -> None:
    """
    Validates the given instance using the given measurement object schema and raises a ValidationError if it is invalid.

    :param instance: the sampledb object
    :param schema: the valid sampledb object schema
    :param path: the path to this subinstance / subschema
    :raise ValidationError: if the schema is invalid.
    """
    if not isinstance(instance, dict):
        raise ValidationError('instance must be dict', path)
    valid_keys = {'_type', 'object_id'}
    required_keys = valid_keys
    schema_keys = set(instance.keys())
    invalid_keys = schema_keys - valid_keys - OPT_IMPORT_KEYS
    if invalid_keys:
        raise ValidationError(f'unexpected keys in schema: {invalid_keys}', path)
    missing_keys = required_keys - schema_keys
    if missing_keys:
        raise ValidationError(f'missing keys in schema: {missing_keys}', path)
    if instance['_type'] != 'measurement':
        raise ValidationError('expected _type "measurement"', path)
    if not isinstance(instance['object_id'], int):
        raise ValidationError('object_id must be int', path)
    if 'component_uuid' in instance and instance['component_uuid'] != flask.current_app.config['FEDERATION_UUID']:
        pass
    elif 'eln_source_url' in instance:
        validate_eln_urls(instance, path)
    else:
        try:
            measurement = objects.get_object(object_id=instance['object_id'])
        except ObjectDoesNotExistError:
            raise ValidationError('object does not exist', path)
        if measurement.action_id is None:
            raise ValidationError('object must be measurement', path)
        action = actions.get_action(measurement.action_id)
        if action.type is None:
            raise ValidationError('object must be measurement', path)
        if ActionType.MEASUREMENT not in {action.type_id, action.type.fed_id}:
            raise ValidationError('object must be measurement', path)


def _validate_user(instance: typing.Dict[str, typing.Any], schema: typing.Dict[str, typing.Any], path: typing.List[str]) -> None:
    """
    Validates the given instance using the given user object schema and raises a ValidationError if it is invalid.

    :param instance: the sampledb object
    :param schema: the valid sampledb object schema
    :param path: the path to this subinstance / subschema
    :raise ValidationError: if the schema is invalid
    """
    if not isinstance(instance, dict):
        raise ValidationError('instance must be dict', path)
    valid_keys = {'_type', 'user_id'}
    required_keys = valid_keys
    schema_keys = set(instance.keys())
    invalid_keys = schema_keys - valid_keys - OPT_IMPORT_KEYS
    if invalid_keys:
        raise ValidationError(f'unexpected keys in schema: {invalid_keys}', path)
    missing_keys = required_keys - schema_keys
    if missing_keys:
        raise ValidationError(f'missing keys in schema: {missing_keys}', path)
    if instance['_type'] != 'user':
        raise ValidationError('expected _type "user"', path)
    if not isinstance(instance['user_id'], int):
        raise ValidationError('user_id must be int', path)
    if instance['user_id'] < 1:
        raise ValidationError('user_id must be positive', path)
    if 'component_uuid' in instance and instance['component_uuid'] != flask.current_app.config['FEDERATION_UUID']:
        pass  # TODO
    elif 'eln_source_url' in instance:
        validate_eln_urls(instance, path)
    else:
        try:
            users.check_user_exists(user_id=instance['user_id'])
        except UserDoesNotExistError:
            raise ValidationError('user does not exist', path)


def _validate_object_reference(instance: typing.Dict[str, typing.Any], schema: typing.Dict[str, typing.Any], path: typing.List[str]) -> None:
    """
    Validates the given instance using the given object reference object schema and raises a ValidationError if it is invalid.

    :param instance: the sampledb object
    :param schema: the valid sampledb object schema
    :param path: the path to this subinstance / subschema
    :raise ValidationError: if the schema is invalid.
    """
    if not isinstance(instance, dict):
        raise ValidationError('instance must be dict', path)
    valid_keys = {'_type', 'object_id'}
    required_keys = valid_keys
    schema_keys = set(instance.keys())
    invalid_keys = schema_keys - valid_keys - OPT_IMPORT_KEYS
    if invalid_keys:
        raise ValidationError(f'unexpected keys in schema: {invalid_keys}', path)
    missing_keys = required_keys - schema_keys
    if missing_keys:
        raise ValidationError(f'missing keys in schema: {missing_keys}', path)
    if instance['_type'] != 'object_reference':
        raise ValidationError('expected _type "object_reference"', path)
    if not isinstance(instance['object_id'], int):
        raise ValidationError('object_id must be int', path)
    if instance['object_id'] < 1:
        raise ValidationError('object_id must be positive', path)
    if 'component_uuid' in instance and instance['component_uuid'] != flask.current_app.config['FEDERATION_UUID']:
        pass
    elif 'eln_source_url' in instance:
        validate_eln_urls(instance, path)
    else:
        try:
            object = objects.get_object(object_id=instance['object_id'])
        except ObjectDoesNotExistError:
            raise ValidationError('object does not exist', path)
        filter_operator = schema.get('filter_operator', 'and')
        action_id_error = None
        if 'action_id' in schema:
            if type(schema['action_id']) is int:
                valid_action_ids = [schema['action_id']]
            else:
                valid_action_ids = schema['action_id']
            if valid_action_ids is not None:
                if object.action_id is None:
                    action_id_error = 'object has no action'
                elif object.action_id not in valid_action_ids:
                    action_id_error = 'object has wrong action'
        if filter_operator == 'and' and action_id_error:
            raise ValidationError(action_id_error, path)
        action_type_id_error = None
        if 'action_type_id' in schema:
            if type(schema['action_type_id']) is int:
                valid_action_type_ids = [schema['action_type_id']]
            else:
                valid_action_type_ids = schema['action_type_id']
            if valid_action_type_ids is not None:
                if object.action_id is None:
                    action_type_id_error = 'object has no action type'
                else:
                    action = actions.get_action(object.action_id)
                    if action.type is None:
                        action_type_id_error = 'object has no action type'
                    elif action.type_id not in valid_action_type_ids and (action.type.fed_id is None or action.type.fed_id >= 0 or action.type.fed_id not in valid_action_type_ids):
                        action_type_id_error = 'object has wrong action type'
        if filter_operator == 'and' and action_type_id_error:
            raise ValidationError(action_type_id_error, path)
        if filter_operator == 'or' and action_id_error and action_type_id_error:
            raise ValidationError(action_id_error, path)


def _validate_plotly_chart(instance: typing.Dict[str, typing.Any], schema: typing.Dict[str, typing.Any], path: typing.List[str]) -> None:
    """
    Validates the given instance using the given text object schema and raises a ValidationError if it is invalid.

    :param instance: the sampledb object
    :param schema: the valid sampledb object schema
    :param path: the path to this subinstance / subschema
    :raise ValidationError: if the schema is invalid.
    """
    if not isinstance(instance, dict):
        raise ValidationError('instance must be dict', path)
    valid_keys = {'_type', 'plotly'}
    required_keys = ['_type', 'plotly']
    schema_keys = instance.keys()
    invalid_keys = schema_keys - valid_keys - OPT_IMPORT_KEYS
    if invalid_keys:
        raise ValidationError(f'unexpected keys in schema: {invalid_keys}', path)
    missing_keys = required_keys - schema_keys
    if missing_keys:
        raise ValidationError(f'missing keys in schema: {missing_keys}', path)
    if instance['_type'] != 'plotly_chart':
        raise ValidationError('expected _type "plotly_chart"', path)
    if isinstance(instance['plotly'], str):
        if len(instance['plotly']) > 0:
            try:
                instance['plotly'] = json.loads(instance['plotly'])
            except json.JSONDecodeError:
                raise ValidationError(_('plotly data must be valid JSON'), path)
        else:
            instance['plotly'] = {}
    if not isinstance(instance['plotly'], dict):
        raise ValidationError(_('plotly must be a dict'), path)

    try:
        plotly.io.from_json(json.dumps(instance['plotly']), 'Figure', False)
    except ValueError:
        raise ValidationError(_('The plotly data must be valid. Look up which schema is supported by plotly.'), path)


def _validate_timeseries(
        instance: typing.Dict[str, typing.Any],
        schema: typing.Dict[str, typing.Any],
        path: typing.List[str]
) -> None:
    """
    Validates the given instance using the given timeseries schema and raises a ValidationError if it is invalid.

    :param instance: the sampledb object
    :param schema: the valid sampledb object schema
    :param path: the path to this subinstance / subschema
    :raise ValidationError: if the schema is invalid.
    """
    if not isinstance(instance, dict):
        raise ValidationError('instance must be dict', path)
    required_keys = {'_type', 'data'}
    valid_keys = required_keys.union({'units', 'dimensionality', 'data'})
    schema_keys = set(instance.keys())
    invalid_keys = schema_keys - valid_keys - OPT_IMPORT_KEYS
    if invalid_keys:
        raise ValidationError(f'unexpected keys in schema: {invalid_keys}', path)
    missing_keys = required_keys - schema_keys
    if missing_keys:
        raise ValidationError(f'missing keys in schema: {missing_keys}', path)
    if instance['_type'] != 'timeseries':
        raise ValidationError('expected _type "timeseries"', path)
    if 'units' not in instance:
        instance['units'] = schema['units']
    if not isinstance(instance['units'], str):
        raise ValidationError('units must be str', path)
    if not units_are_valid(instance['units']):
        raise ValidationError('Invalid/Unknown units', path)
    try:
        if isinstance(schema['units'], str):
            schema_units = schema['units']
        else:
            schema_units = schema['units'][0]
        dimensionality_from_schema_units = get_dimensionality_for_units(schema_units)
    except Exception:
        raise ValidationError('Unable to determine dimensionality', path)
    if isinstance(schema['units'], str) and instance['units'] != schema['units']:
        if dimensionality_from_schema_units != get_dimensionality_for_units(instance['units']):
            raise ValidationError(f'Invalid units, expected {schema["units"]}', path)
    if isinstance(schema['units'], list) and instance['units'] not in schema['units']:
        if dimensionality_from_schema_units != get_dimensionality_for_units(instance['units']):
            raise ValidationError(f'Invalid units, expected one of {", ".join(schema["units"])}', path)
    if 'dimensionality' in instance:
        if not isinstance(instance['dimensionality'], str):
            raise ValidationError('dimensionality must be str', path)
        if instance['dimensionality'] != dimensionality_from_schema_units:
            raise ValidationError('dimensionality must match units', path)
    else:
        instance['dimensionality'] = dimensionality_from_schema_units

    if not isinstance(instance['data'], list):
        raise ValidationError('data must be list', path)

    is_relative_time = len(instance['data']) > 0 and len(instance['data'][0]) > 0 and type(instance['data'][0][0]) in (int, float)
    if (
        (not is_relative_time and not all(isinstance(entry, (list, tuple)) and len(entry) in (2, 3) and type(entry[0]) is str and all(type(value) in (int, float) for value in entry[1:]) for entry in instance['data'])) or
        (is_relative_time and not all(isinstance(entry, (list, tuple)) and len(entry) in (2, 3) and type(entry[0]) in (int, float) and all(type(value) in (int, float) for value in entry[1:]) for entry in instance['data']))
    ):
        raise ValidationError('data must be list of lists containing either a datetime string or relative time in seconds, and 1 or 2 numbers', path)

    existing_times = set()
    for i, entry in enumerate(instance['data']):
        time, magnitude = entry[:2]
        if not math.isfinite(magnitude):
            raise ValidationError('magnitude must be finite', path)
        if time in existing_times:
            raise ValidationError('duplicate point in timeseries', path)
        if not is_relative_time:
            try:
                datetime.datetime.strptime(time, '%Y-%m-%d %H:%M:%S.%f')
            except Exception:
                raise ValidationError('invalid datetime in timeseries, expected format: YYYY-MM-DD hh:mm:ss.ffffff', path)
        elif not math.isfinite(time):
            raise ValidationError('relative time must be finite', path)
        existing_times.add(time)
        calculated_magnitude_in_base_units = get_magnitude_in_base_units(magnitude=decimal.Decimal(magnitude), units=instance['units'])
        if len(entry) == 3:
            magnitude_in_base_units = entry[2]
            if not math.isfinite(magnitude_in_base_units):
                raise ValidationError('magnitude_in_base_units must be finite', path)
            if not math.isclose(float(calculated_magnitude_in_base_units), magnitude_in_base_units):
                raise ValidationError('magnitude_in_base_units and magnitude do not match', path)
        else:
            instance['data'][i] = [time, float(magnitude), float(calculated_magnitude_in_base_units)]


def _validate_file(
        instance: typing.Dict[str, typing.Any],
        schema: typing.Dict[str, typing.Any],
        path: typing.List[str],
        file_names_by_id: typing.Optional[typing.Dict[int, str]] = None
) -> None:
    """
    Validates the given instance using the given file object schema and raises a ValidationError if it is invalid.

    :param instance: the sampledb file object
    :param schema: the valid sampledb file object schema
    :param path: the path to this subinstance / subschema
    :param file_names_by_id: a dict mapping file IDs to file names, or None
    :raise ValidationError: if the schema is invalid
    """
    if not isinstance(instance, dict):
        raise ValidationError('instance must be dict', path)
    valid_keys = {'_type', 'file_id'}
    required_keys = valid_keys
    schema_keys = set(instance.keys())
    invalid_keys = schema_keys - valid_keys - OPT_IMPORT_KEYS
    if invalid_keys:
        raise ValidationError(f'unexpected keys in schema: {invalid_keys}', path)
    missing_keys = required_keys - schema_keys
    if missing_keys:
        raise ValidationError(f'missing keys in schema: {missing_keys}', path)
    if instance['_type'] != 'file':
        raise ValidationError('expected _type "file"', path)
    if not isinstance(instance['file_id'], int):
        raise ValidationError('file_id must be int', path)
    if file_names_by_id is not None:
        if instance['file_id'] not in file_names_by_id:
            raise ValidationError('file does not exist', path)
        file_name = file_names_by_id[instance['file_id']]
        if 'extensions' in schema:
            for extension in schema['extensions']:
                if file_name.lower().endswith(extension.lower()):
                    break
            else:
                raise ValidationError(f'file name should have one of these extensions: {", ".join(schema["extensions"])}', path)


def validate_eln_urls(
        instance: typing.Dict[str, typing.Any],
        path: typing.List[str]
) -> None:
    for key in ['eln_source_url', 'eln_user_url', 'eln_object_url']:
        if instance.get(key) is not None:
            try:
                parse_url(instance[key], valid_schemes=('http', 'https'))
            except InvalidURLError:
                raise ValidationError(f'{key} must be a valid http oder https URL', path)



# coding: utf-8
"""
Tags or keywords are a method of organizing object. Internally, tags are part
of an objects data as text entries. In addition, however, the list of tags and
the number of objects each tag has been used for is stored in the database, so
that users can get tag suggestions and see which tags are used more to foster
a more unified vocabulary without strictly enforcing a whitelist of tags.
"""


import dataclasses
import typing
from .. import db
from ..models import tags, objects


@dataclasses.dataclass(frozen=True)
class Tag:
    """
    This class provides an immutable wrapper around models.tags.Tag.
    """
    id: int
    name: str
    uses: int

    @classmethod
    def from_database(cls, tag: tags.Tag) -> 'Tag':
        return Tag(id=tag.id, name=tag.name, uses=tag.uses)


def get_object_tags(object: objects.Object) -> typing.Sequence[str]:
    if object.data is not None and 'tags' in object.data:
        tags_entry = object.data['tags']
        if isinstance(tags_entry, dict) and '_type' in tags_entry and tags_entry['_type'] == 'tags' and 'tags' in tags_entry:
            return [tag.strip() for tag in tags_entry["tags"]]
    return ()


def update_object_tag_usage(object: objects.Object, new_subversion: bool = False) -> None:
    previous_tags: typing.Sequence[str] = ()
    new_tags = get_object_tags(object)
    if object.component_id is None or object.fed_object_id is None or object.fed_version_id is None:
        if object.version_id > 0:
            previous_object = objects.Objects.get_object_version(object.object_id, object.version_id - 1)
            if previous_object is not None:
                previous_tags = get_object_tags(previous_object)
    else:
        current = objects.Objects.get_current_fed_object(object.component_id, object.fed_object_id)
        if current is None:
            return
        if current.version_id != object.version_id:
            return
        if new_subversion:
            previous_object = objects.Objects.get_previous_subversion(object.object_id, object.version_id)
            if previous_object is not None:
                previous_tags = get_object_tags(previous_object)
        elif object.fed_version_id > 0:
            previous_object = None
            prev_version = object.fed_version_id - 1
            while previous_object is None and prev_version >= 0:
                previous_object = objects.Objects.get_fed_object_version(object.component_id, object.fed_object_id, prev_version)
                prev_version -= 1
            if previous_object is not None:
                previous_tags = get_object_tags(previous_object)
    removed_tags = set(previous_tags) - set(new_tags)
    added_tags = set(new_tags) - set(previous_tags)
    for tag_name in added_tags:
        tag = tags.Tag.query.filter_by(name=tag_name).first()
        if tag is not None:
            tag.uses += 1
        else:
            tag = tags.Tag(name=tag_name, uses=1)
        db.session.add(tag)
    for tag_name in removed_tags:
        tag = tags.Tag.query.filter_by(name=tag_name).first()
        if tag is not None:
            tag.uses = max(0, tag.uses - 1)
            if tag.uses > 0:
                db.session.add(tag)
            else:
                db.session.delete(tag)
    db.session.commit()


def get_tags() -> typing.Sequence[Tag]:
    return [Tag.from_database(tag) for tag in tags.Tag.query.order_by(tags.Tag.uses).all()]



# coding: utf-8
"""
Drop the description column from the action_types table.
"""

import flask_sqlalchemy

from .utils import table_has_column


def run(db: flask_sqlalchemy.SQLAlchemy) -> bool:
    # Skip migration by condition
    if not table_has_column('action_types', 'description'):
        return False

    # Perform migration
    db.session.execute(db.text("""
            ALTER TABLE action_types
            DROP COLUMN description
        """))

    return True



# coding: utf-8
"""

"""
import datetime
import typing

from sqlalchemy.orm import Query, Mapped

from .. import db
from .utils import Model


class Component(Model):
    __tablename__ = 'components'

    id: Mapped[int] = db.Column(db.Integer, primary_key=True)
    address: Mapped[typing.Optional[str]] = db.Column(db.Text, nullable=True)
    uuid: Mapped[str] = db.Column(db.Text, nullable=False, unique=True)
    name: Mapped[typing.Optional[str]] = db.Column(db.Text, nullable=True, unique=True)
    description: Mapped[str] = db.Column(db.Text, nullable=False, default='')
    last_sync_timestamp: Mapped[typing.Optional[datetime.datetime]] = db.Column(db.TIMESTAMP(timezone=True), nullable=True)
    discoverable: Mapped[bool] = db.Column(db.Boolean, nullable=False, default=True, server_default=db.true())
    fed_login_available: Mapped[bool] = db.Column(db.Boolean, nullable=False, default=False, server_default=db.false())

    if typing.TYPE_CHECKING:
        query: typing.ClassVar[Query["Component"]]

    def __init__(
            self,
            uuid: str,
            name: typing.Optional[str] = None,
            description: typing.Optional[str] = '',
            address: typing.Optional[str] = None,
            last_sync_timestamp: typing.Optional[datetime.datetime] = None,
            discoverable: bool = True,
            fed_login_available: bool = False
    ) -> None:
        super().__init__(
            address=address,
            uuid=uuid,
            name=name,
            description=description,
            last_sync_timestamp=last_sync_timestamp,
            discoverable=discoverable,
            fed_login_available=fed_login_available
        )

    def __repr__(self) -> str:
        return f'<{type(self).__name__}(id={self.id}, address={self.address}, uuid={self.uuid}, name={self.name}, description={self.description}, fed_login_available={self.fed_login_available})>'


class ComponentInfo(Model):
    __tablename__ = 'component_infos'

    id: Mapped[int] = db.Column(db.Integer, primary_key=True)
    uuid: Mapped[str] = db.Column(db.Text, nullable=False)
    source_uuid: Mapped[str] = db.Column(db.Text, nullable=False)
    address: Mapped[typing.Optional[str]] = db.Column(db.Text, nullable=True)
    name: Mapped[typing.Optional[str]] = db.Column(db.Text, nullable=True)
    discoverable: Mapped[bool] = db.Column(db.Boolean, nullable=False)
    distance: Mapped[int] = db.Column(db.Integer, nullable=False)

    if typing.TYPE_CHECKING:
        query: typing.ClassVar[Query["ComponentInfo"]]

    def __repr__(self) -> str:
        return f'<{type(self).__name__}(id={self.id}, address={self.address}, uuid={self.uuid}, name={self.name}, source_uuid={self.source_uuid}, distance={self.distance})>'



# coding: utf-8
"""
Create default action type for schema templates.
"""

import flask_sqlalchemy

from ..actions import ActionType
from ..languages import Language
from .utils import table_has_column


def run(db: flask_sqlalchemy.SQLAlchemy) -> bool:
    # Add TEMPLATE to default action types
    existing_action_type_ids = [
        action_type[0]
        for action_type in db.session.execute(db.text("""
                    SELECT id
                    FROM action_types;
                """)).fetchall()
    ]
    if ActionType.TEMPLATE in existing_action_type_ids:
        return False

    action_type_template = {
        'id': ActionType.TEMPLATE,
        'admin_only': False,
        'show_on_frontpage': False,
        'show_in_navbar': True,
        'enable_labels': True,
        'enable_files': True,
        'enable_locations': True,
        'enable_publications': True,
        'enable_comments': True,
        'enable_activity_log': True,
        'enable_related_objects': True,
        'enable_project_link': False,
        'disable_create_objects': True,
        'is_template': True
    }

    db.session.execute(db.text("""
      INSERT INTO action_types (id, admin_only, show_on_frontpage, show_in_navbar, enable_labels, enable_files, enable_locations, enable_publications, enable_comments, enable_activity_log, enable_related_objects, disable_create_objects, is_template)
      VALUES (:id, :admin_only, :show_on_frontpage, :show_in_navbar, :enable_labels, :enable_files, :enable_locations, :enable_publications, :enable_comments, :enable_activity_log, :enable_related_objects, :disable_create_objects, :is_template)
  """), params=action_type_template)
    performed_migration_type = True

    if table_has_column('action_types', 'show_in_object_filters'):
        # Hide schema template action type per default
        db.session.execute(db.text("""
            UPDATE action_types
            SET show_in_object_filters = FALSE
            WHERE id = :id
        """), params={'id': ActionType.TEMPLATE})

    # Add translations for TEMPLATE action type
    action_type_template_translations = [
        {
            'action_type_id': ActionType.TEMPLATE,
            'language_id': Language.ENGLISH,
            'name': 'Schema Template',
            'description': 'These Actions represent schema templates that can be included into other actions.',
            'object_name': '-',
            'object_name_plural': '-',
            'view_text': '-',
            'perform_text': '-'
        },
        {
            'action_type_id': ActionType.TEMPLATE,
            'language_id': Language.GERMAN,
            'name': 'Schema-Vorlage',
            'description': 'Aktionen dieses Typs können in andere Aktionen als Vorlagen für Schemata eingebunden werden.',
            'object_name': '-',
            'object_name_plural': '-',
            'view_text': '-',
            'perform_text': '-'
        }
    ]

    performed_migration_translation = False

    for template_translation in action_type_template_translations:
        db.session.execute(db.text("""
         INSERT INTO action_type_translations (action_type_id, language_id, name, description, object_name, object_name_plural, view_text, perform_text)
          VALUES (:action_type_id, :language_id,:name, :description, :object_name, :object_name_plural, :view_text, :perform_text)
      """), params=template_translation)
        performed_migration_translation = True

    return performed_migration_type and performed_migration_translation



# coding: utf-8
"""
Replace the view user_object_permissions_by_all to add a
requires_instruments column to allow filtering based on the
DISABLE_INSTRUMENTS configuration value.
"""

import flask_sqlalchemy


def run(db: flask_sqlalchemy.SQLAlchemy) -> bool:
    # Perform migration
    db.session.execute(db.text("""
    DROP VIEW IF EXISTS user_object_permissions_by_all
    """))
    db.session.execute(db.text("""
    CREATE VIEW user_object_permissions_by_all
    AS SELECT
    u.user_id AS user_id,
    u.object_id AS object_id,
    MAX(u.permissions_int) AS permissions_int,
    u.requires_anonymous_users AS requires_anonymous_users,
    u.requires_instruments AS requires_instruments
    FROM (
        SELECT
        NULL::integer AS user_id,
        object_id AS object_id,
        CASE all_user_object_permissions.permissions
            WHEN 'NONE' THEN 0
            WHEN 'READ' THEN 1
            WHEN 'WRITE' THEN 2
            WHEN 'GRANT' THEN 3
            ELSE 0
        END AS permissions_int,
        FALSE AS requires_anonymous_users,
        FALSE AS requires_instruments
        FROM all_user_object_permissions
    UNION
        SELECT
        NULL::integer AS user_id,
        object_id AS object_id,
        CASE anonymous_user_object_permissions.permissions
            WHEN 'NONE' THEN 0
            WHEN 'READ' THEN 1
            WHEN 'WRITE' THEN 2
            WHEN 'GRANT' THEN 3
            ELSE 0
        END AS permissions_int,
        TRUE AS requires_anonymous_users,
        FALSE AS requires_instruments
        FROM anonymous_user_object_permissions
    UNION
        SELECT
        association.user_id AS user_id,
        objects.object_id AS object_id,
        3 AS permissions_int,
        FALSE AS requires_anonymous_users,
        TRUE AS requires_instruments
        FROM association AS association
        JOIN actions AS actions ON association.instrument_id = actions.instrument_id
        JOIN objects_current AS objects ON objects.action_id = actions.id
    UNION
        SELECT
        user_object_permissions.user_id AS user_id,
        user_object_permissions.object_id AS object_id,
        CASE user_object_permissions.permissions
            WHEN 'NONE' THEN 0
            WHEN 'READ' THEN 1
            WHEN 'WRITE' THEN 2
            WHEN 'GRANT' THEN 3
            ELSE 0
        END AS permissions_int,
        FALSE AS requires_anonymous_users,
        FALSE AS requires_instruments
        FROM user_object_permissions
    UNION
        SELECT
        user_group_memberships.user_id AS user_id,
        group_object_permissions.object_id AS object_id,
        CASE group_object_permissions.permissions
            WHEN 'NONE' THEN 0
            WHEN 'READ' THEN 1
            WHEN 'WRITE' THEN 2
            WHEN 'GRANT' THEN 3
            ELSE 0
        END AS permissions_int,
        FALSE AS requires_anonymous_users,
        FALSE AS requires_instruments
        FROM user_group_memberships AS user_group_memberships
        JOIN group_object_permissions AS group_object_permissions ON user_group_memberships.group_id = group_object_permissions.group_id
    UNION
        SELECT
        user_project_permissions.user_id AS user_id,
        project_object_permissions.object_id AS object_id,
        LEAST(
            CASE project_object_permissions.permissions
                WHEN 'NONE' THEN 0
                WHEN 'READ' THEN 1
                WHEN 'WRITE' THEN 2
                WHEN 'GRANT' THEN 3
                ELSE 0
            END,
            CASE user_project_permissions.permissions
                WHEN 'NONE' THEN 0
                WHEN 'READ' THEN 1
                WHEN 'WRITE' THEN 2
                WHEN 'GRANT' THEN 3
                ELSE 0
            END
        ) AS permissions_int,
        FALSE AS requires_anonymous_users,
        FALSE AS requires_instruments
        FROM user_project_permissions AS user_project_permissions
        JOIN project_object_permissions AS project_object_permissions ON user_project_permissions.project_id = project_object_permissions.project_id
    UNION
        SELECT
        user_group_memberships.user_id AS user_id,
        project_object_permissions.object_id AS object_id,
        LEAST(
            CASE project_object_permissions.permissions
                WHEN 'NONE' THEN 0
                WHEN 'READ' THEN 1
                WHEN 'WRITE' THEN 2
                WHEN 'GRANT' THEN 3
                ELSE 0
            END,
            CASE group_project_permissions.permissions
                WHEN 'NONE' THEN 0
                WHEN 'READ' THEN 1
                WHEN 'WRITE' THEN 2
                WHEN 'GRANT' THEN 3
                ELSE 0
            END
        ) AS permissions_int,
        FALSE AS requires_anonymous_users,
        FALSE AS requires_instruments
        FROM user_group_memberships AS user_group_memberships
        JOIN group_project_permissions ON group_project_permissions.group_id = user_group_memberships.group_id
        JOIN project_object_permissions AS project_object_permissions ON group_project_permissions.project_id = project_object_permissions.project_id
    ) AS u
    GROUP BY (u.object_id, u.user_id, u.requires_anonymous_users, u.requires_instruments)
    """))
    return True



from . import automatic_schema_updates
from . import files
from . import forms
from . import objects
from . import object_form
from . import object_form_parser
from . import permissions
from . import search
from . import versions
from . import view


__all__ = [
    'automatic_schema_updates',
    'files',
    'forms',
    'objects',
    'object_form',
    'object_form_parser',
    'permissions',
    'search',
    'versions',
    'view',
]



# coding: utf-8
"""
Script for updating an action in SampleDB.

Usage: sampledb update_action <action_id> <name> <description> <schema_file_name>
"""

import json
import sys
import typing

from .. import create_app
from ..logic.actions import check_action_exists
from ..logic.action_translations import set_action_translation
from ..logic.schemas import validate_schema
from ..logic.errors import ActionDoesNotExistError, ValidationError
from ..logic.languages import Language


def main(arguments: typing.List[str]) -> None:
    if len(arguments) != 4:
        print(__doc__)
        sys.exit(1)
    action_id_str, name, description, schema_file_name = arguments
    try:
        action_id = int(action_id_str)
    except ValueError:
        print("Error: action_id must be an integer", file=sys.stderr)
        sys.exit(1)

    app = create_app()
    with app.app_context():
        try:
            check_action_exists(action_id)
        except ActionDoesNotExistError:
            print('Error: no action with this id exists', file=sys.stderr)
            sys.exit(1)
        with open(schema_file_name, 'r', encoding='utf-8') as schema_file:
            schema = json.load(schema_file)
        try:
            validate_schema(schema)
        except ValidationError as e:
            print(f'Error: invalid schema: {str(e)}', file=sys.stderr)
            sys.exit(1)
        set_action_translation(
            language_id=Language.ENGLISH,
            action_id=action_id,
            name=name,
            description=description
        )
        print("Success: the action has been updated in SampleDB")
